/*============================================================================
 * adc.c
 *      Module to support reading analog inputs.  Used in this project for
 *      reading battery health.
 * Author: Jon W. Wallace 
 *==========================================================================*/

#include <inttypes.h>
#include <xc.h>
#include "adc.h"
#include "timer.h"
#include "timer_ms.h"

#define ADC_SOURCE_NONE         0x00
#define ADC_SOURCE_MAIN_BATT    0x01
#define ADC_SOURCE_COIN_CELL    0x02

#define ADC_PORT_TRIS   (TRISB)
#define ADC_PORT_ANSEL  (ANSELB)
#define ADC_PORT_BIT    (1 << 7)  // RB7 Pin 27  

#define ADCON0_ON       (1 << 7)
#define ADCON0_CS       (1 << 4)
#define ADCON0_GO       (1 << 0)

#define ADREF_FVR       0x03
#define ADREF_VDD       0x00

#define ADCLK_CS_DIV16  3

#define PRE_TIME        0x0
#define ACQ_TIME        0x0

// ADPCH
// ADRESH
// ADRESL
// Conversion clock, can use dedicated FRC
// CS(ADCON0<4>) = 1?

// FVR (better than using VDD)
// Set FVREN in FVRCON
// ADFVR configure gain amps
//

#define FVRCON_FVRRDY    (1 << 6)
#define FVRCON_FVREN     (1 << 7)
#define FVRCON_TSEN      (1 << 5)
#define FVRCON_ADFVR_4x  0x03            // 4.096V reference
#define FVRCON_ADFVR_2x  0x02            // 2.048V reference
#define FVRCON_ADFVR_1x  0x01            // 1.024V
#define FVRCON_ADFVR_OFF 0x00

void adc_init()
{
    // Set to be analog input (uncomment if used)
    ADC_PORT_TRIS |= ADC_PORT_BIT;
    ADC_PORT_ANSEL |= ADC_PORT_BIT;

    FVRCON = FVRCON_FVREN | FVRCON_ADFVR_2x | FVRCON_TSEN;
    while (!(FVRCON & FVRCON_FVRRDY));
    
    ADPCH = ADPCH_MAIN_BATT;

    ADPREL = (PRE_TIME & 0x00ff);
    ADPREH = PRE_TIME >> 8;
    
    ADACQL = (ACQ_TIME & 0x00ff);
    ADACQH = ACQ_TIME >> 8;

    ADCAP = 0;          // No additional capacitance

    ADRPT = 0;
    ADACT = 0;

    ADCON1 = 0;
    ADCON2 = 0;
    ADCON3 = 0;

    ADCLK = 0;
    //ADCLK = ADCLK_CS_DIV16;
    ADREF = ADREF_FVR;
    //ADREF = ADREF_VDD;
    ADCON0 = ADCON0_ON | ADCON0_CS;
}

void adc_start()
{
    // Start conversion
    ADCON0 |= ADCON0_GO;

}

void adc_stop()
{
    // Stop conversion
    ADCON0 &= ~ADCON0_GO;
}

uint8_t adc_ready()
{
    return((ADCON0 & ADCON0_GO) == 0);
}

uint16_t adc_read()
{
    return((ADRESH << 8) | ADRESL);
}

// adc_set_source
//      Sets what the ADC is connected to.
void adc_set_source(uint8_t source)
{
    if (ADPCH != source)
    {
        ADPCH = source;
        timer_ms_delay(1);
    }
}

    
void adc_sleep()
{
    // Turn off ADC
    ADCON0 = 0;
    
    // Turn off analog function of pin
    ADC_PORT_ANSEL &= ~ADC_PORT_BIT;
    FVRCON = 0;

    ADPCH = 0x0;

    ADPREL = 0;
    ADPREH = 0;
    
    ADACQL = 0;
    ADACQH = 0;

    ADCAP = 0; 
    ADRPT = 0;
    ADACT = 0;

    ADCON1 = 0;
    ADCON2 = 0;
    ADCON3 = 0;

    ADCLK = 0;
    ADREF = ADREF_VDD;
}
