/*============================================================================
 * battery.c
 *      Code to read and scale battery voltage read via ADC.
 * Author: Jon W. Wallace 
 *==========================================================================*/

#define BATT_DIV_R1             10000L
#define BATT_DIV_R2             220000L
#define BATT_DIODE_DROP         700L

//#define VREF_MV 1024
#define VREF_MV 2048L
//#define VREF_MV 4096

#define N_AVG   16

#include <inttypes.h>
#include "battery.h"
#include "adc.h"
#include "timer_ms.h"
#include "config.h"
#include "uart.h"

//#define DEBUG_BATTERY_THRESHOLD

uint32_t get_avg_adc()
{
    uint32_t val;
    uint8_t i;
    
    val = 0;
    for (i=0; i<N_AVG; i++)
    {
        adc_start();
        while (!adc_ready());
        val += (uint32_t) adc_read();
    }
    val = val/N_AVG;

    return(val);
}

uint16_t battery_main_mv()
{
    uint32_t v_batt_mv, val;
    
    // Select source 
    adc_set_source(ADPCH_MAIN_BATT);
    
    // Read ADC
    val = get_avg_adc();

    //uart_write_dec_int(val);
    //uart_write_nl();

    // Order of operations is important here to avoid UINT32 overflow.
    v_batt_mv = (uint16_t)((BATT_DIV_R1 + BATT_DIV_R2)/BATT_DIV_R1*(uint32_t)val*VREF_MV/65536L + BATT_DIODE_DROP);
    return(v_batt_mv);
}

uint16_t battery_coin_mv()
{
    uint32_t v_batt_mv, val;
    
    // Select source 
    adc_set_source(ADPCH_VBATT);
    
    // Read ADC
    val = get_avg_adc();

    //uart_write_dec_int(val);
    //uart_write_nl();
     
    //v_batt_mv = val*VREF_MV*3L/65536;
    /* I did not get good accuracy assuming Vbatt/3 is what is actually measured.  I measured the following
     * voltage to ADC value pairs:
     * V        ADC
     * 1        8326
     * 2        15521
     * 3        23480
     * 4        31416
     * 5        39876
     * Linear fit gives adc = v*7900 + 25  (can ignore the 25)
     */
    v_batt_mv = val*1000L/7900L;
    return(v_batt_mv);
}

void battery_check(void)
{
    uint32_t coin_mv;
    uint32_t main_mv;

    main_mv = battery_main_mv();

    uart_write_string("Main Battery: ");
    uart_write_dec_int(main_mv);
    uart_write_string(" mv\r\n");

    coin_mv = battery_coin_mv();

    uart_write_string("Coin Cell: ");
    uart_write_dec_int(coin_mv);
    uart_write_string(" mv\r\n");

    if (coin_mv < 1000)
    {
        uart_write_string("WARNING: Coin cell dead or missing!\r\n");
    }
    else if (coin_mv < 2500)
    {
        uart_write_string("WARNING: Coin cell critically low!\r\n");
    }
    else if (coin_mv < 2750)
    {
        uart_write_string("WARNING: Coin cell low.\r\n");
    }

}

void battery_main_print()
{
    uint8_t main_mv;
    uint8_t digit;
    
    main_mv = (battery_main_mv()/100);

    digit = main_mv / 100;
    main_mv = main_mv % 100;
    uart_write(digit + '0');
    
    digit = main_mv / 10;
    main_mv = main_mv % 10;
    uart_write(digit + '0');

    uart_write('.');
    
    digit = main_mv;
    uart_write(digit + '0');

    uart_write('V');
}

void battery_sleep()
{
    uint16_t battery_on_thres;

    // Make sure main battery monitor pin is set as source
    adc_set_source(ADPCH_MAIN_BATT);

    // Store the wakeup voltage required to come out of sleep
    battery_on_thres = system_config.batt_on_threshold;

    //(uint16_t)((uint32_t)system_config.batt_on_threshold*(uint32_t)(BATT_DIV_R1)*65536L/(BATT_DIV_R1 + BATT_DIV_R2)/VREF_MV);

    // Store in registers that are persistent across sleeps.
    VB1GPR = battery_on_thres >> 8;
    VB0GPR = battery_on_thres & 0xff;
}

// battery_sleep_check
//      This function checks if we should wakeup from sleep based on the battery health.
//      If battery looks OK, we return 1, otherwise return 0.
uint8_t battery_sleep_wakeup_ok()
{
    uint16_t battery_on_thres;
    uint16_t battery_mv;

    battery_on_thres = ((uint16_t)VB1GPR << 8) | (uint16_t)VB0GPR;
    battery_mv = battery_main_mv();

#ifdef DEBUG_BATTERY_THRESHOLD    
    uart_write_string("Required battery: ");
    uart_write_dec_int(battery_on_thres);
    uart_write_nl();
    uart_write_nl();

    uart_write_string("Battery check: ");
    uart_write_dec_int(battery_mv);
    uart_write_nl();
    uart_write_nl();
#endif
    
    if (battery_mv > battery_on_thres)
    {
        return(1);
    }

    return(0);
}
