/*============================================================================
 * command.c
 *      Command shell for the RFID board.
 * Author: Jon W. Wallace 
 *==========================================================================*/

#include "global.h"
#include "uart.h"
#include "rtc_timer.h"
#include "config.h"
#include "test.h"
#include "tag_log.h"

uart_char input_string[64];
uint8_t byte_buffer[16];
uint32_t word_buffer[16];

void command_init()
{
}


void command_error()
{
    uart_write_string("???\r\n");
}

void config_print_help()
{
    uart_write_string("Commands:\r\n");
    uart_write_string("?  Help\r\n");
    uart_write_string("a  Set auto, poll, delay\r\n");
    uart_write_string("b  Set battery thresholds\r\n");
    uart_write_string("c  Set cache parameters\r\n");
    uart_write_string("d  Set RTC date\r\n");
    uart_write_string("i  Set board ID\r\n");
    uart_write_string("p  Print config\r\n");
    uart_write_string("r  Read config from EEPROM\r\n");
    uart_write_string("s  Set antenna switch parameters\r\n");
    uart_write_string("t  Set RTC time\r\n");
    uart_write_string("w  Write config to EEPROM\r\n");
    uart_write_string("q  Quit\r\n");
}

void config_panel()
{
    uint8_t quit;
    uint8_t b;
    
    quit = 0;
    while (!quit)
    {
        // Set timeout for autostart
        if (system_config.time_auto) uart_set_timeout(system_config.time_auto*1000);
        
        uart_write_prompt("conf> ");
        uart_read_string(input_string, sizeof(input_string), UART_PROMPT_ECHO);

        switch(input_string[0])
        {
            case UART_TIMEOUT_CHAR:
                uart_write_string("Timeout.\r\n");
            case 'q':
                quit = 1;
                break;

            case 'a':
                config_set_times();
                break;

            case 'b':
                config_set_battery();
                break;

            case 'c':
                config_set_cache();
                break;

            case 'd':
                // Set the date
                rtc_user_set_date();
                break;

            case 'i':
                config_set_board_id();
                break;

            case 'p':
                config_print();
                break;

            case 'r':
                switch (config_load())
                {
                    case -1:
                        uart_write_string("Read failed! Using default.\n");
                        config_default();
                        break;

                    case -2:
                        uart_write_string("Checksum error! Using default.\n");
                        config_default();
                        break;
                }
                break;

            case 's':
                config_set_antenna_switch();
                break;

            case 't':
                // Set the time
                rtc_user_set_time();
                break;
                
                
            case 'w':
                if (config_save() != 0)
                {
                    uart_write_string("Config write failed!\n");
                    break;
                }
                break;

                
            case '?':
                config_print_help();
                break;

            default:
                command_error();
                break;
        }
    }
}

void command_print_version()
{
    uart_write_string("Lafayette Switched LF RFID Reader\r\n");
    uart_write_string("Version: ");
    uart_write_string(HW_VERSION_STRING);
    uart_write_string(" ");
    uart_write_string(SW_VERSION_STRING);
    uart_write_nl();
    uart_write_string("Board ID: ");
    uart_write_string(system_config.board_id);
    uart_write_nl();
    uart_write_string("Current Date: ");
    rtc_print_date_time(&rtc_time, 0);
    uart_write_nl();
}

void command_print_help()
{
    uart_write_string("Commands:\r\n");
    uart_write_string("?  Help\r\n");
    uart_write_string("c  Configuration\r\n");
    uart_write_string("p  Print tag log\r\n");
    uart_write_string("r  Run RFID logging\r\n");
    uart_write_string("t  Test Console\r\n");
    uart_write_string("v  Version\r\n");
    uart_write_string("E  Erase tag log\r\n");
}

void command_run()
{
    uint16_t n, i;
    
    command_print_version();

    while (1)
    {
        // Set timeout for autostart
        if (system_config.time_auto) uart_set_timeout(system_config.time_auto*1000);
        
        uart_write_prompt("cmd> ");
        uart_read_string(input_string, sizeof(input_string), UART_PROMPT_ECHO);

        switch(input_string[0])
        {
            case 'c':
                config_panel();
                break;

            case 'p':
                tag_log_dump_formatted();
                break;
                
            case 'P':
                tag_log_dump();
                break;

            case 'E':
                uart_write_prompt("Tag log will be erased.  Are you sure (y/n)? ");
                uart_read_string(input_string, sizeof(input_string), UART_PROMPT_ECHO);
                if (input_string[0] == 'y')
                {
                    if (tag_log_erase() == 0)
                        uart_write_string("Erased.\r\n");
                    else
                        uart_write_string("Erase error!\r\n");
                }
                break;
            
            case 'v':
                command_print_version();
                break;

            case UART_TIMEOUT_CHAR:  // Timeout -> autostart.
                uart_write_string("\r\nAutomatic start.\r\n");
            case 'r':
                tag_log_run();
                break;

            case 't':
                test_panel();
                break;

            case '?':
                command_print_help();
                break;

            default:
                command_error();
                break;
        }
    }
}

