/*============================================================================
 * config.c
 *      Code to support storing configuration of the board in EEPROM
 *      and reloading when the board comes up.
 * Author: Jon W. Wallace 
 *==========================================================================*/

#include <inttypes.h>
#include <string.h>

#include "global.h"
#include "config.h"
#include "eeprom.h"
#include "uart.h"
#include "rtc_timer.h"
#include "command.h"
#include "parse.h"


config_def system_config;

void config_init()
{
    memset(&system_config, 0, sizeof(system_config));
}

uint16_t config_checksum()
{
    uint16_t *p;
    uint16_t check;
    uint16_t i;

    check = 0;
    p = (uint16_t *) &system_config + 1;
    for (i=0; i<(sizeof(system_config)/sizeof(uint16_t)-1); i++)
    {
        check += *p;
        p++;
    }

    return(check);
}

// Save configuration from RAM to EEPROM
int8_t config_save()
{
    uint8_t *p;
    uint8_t b1;
    uint16_t i;

    system_config.checksum = config_checksum();
    
    // Write configuration data to board.  It is not much data, so let's
    // be conservative and just update bytes that need to be changed.
    p = (uint8_t *) &system_config;
    for (i=0; i<sizeof(system_config); i++)
    {
        if (eeprom_read_bytes(EEPROM_CONFIG_BASE + i, &b1, 1) != 0)
        {
            return(-1);
        }
        if (b1 != *p)
        {
            if (eeprom_write_bytes(EEPROM_CONFIG_BASE + i, p, 1) != 0)
            {
                return(-1);
            }
        }
        p++;
    }
    return(0);
}

// Load configuration items into RAM
int8_t config_load()
{
    uint8_t *p;
    uint16_t i, count;
    uint16_t check;
    
    check = 0;
    p = (uint8_t *) &system_config;
    for (i=0; i<sizeof(system_config); )
    {
        count = sizeof(system_config)-i;
        if (count > 128) count=128;

        if (eeprom_read_bytes(EEPROM_CONFIG_BASE, p, count) < 0) return(-1);

        i += count;
        p += count;
    }

    if (config_checksum() != system_config.checksum)
    {
        return(-2);
    }
    
    return(0);
}

void config_default()
{
    memset(&system_config, 0, sizeof(system_config));

    system_config.time_auto = 0;
    system_config.time_poll = 500;
    system_config.time_delay = 100;

    system_config.batt_off_threshold = 11000;
    system_config.batt_on_threshold = 12000;
        
    system_config.cache_size = 16;
    system_config.cache_timeout = 300;

    system_config.ant_start = 0;
    system_config.ant_stop = 0;
}

void config_set_board_id()
{
    uart_write_prompt("New board ID: ");
    uart_read_string(system_config.board_id, sizeof(system_config.board_id), UART_PROMPT_ECHO);
}

void config_set_times()
{
    uint32_t input;
    uint16_t elem;
    
    uart_write_string("Auto start (s): ");
    uart_write_dec_int(system_config.time_auto);
    uart_write_nl();
    
    uart_write_prompt("New auto start (must be 0s=OFF or >=5s):  ");
    uart_read_string(input_string, sizeof(input_string), UART_PROMPT_ECHO);
    if (input_string[0] == UART_TIMEOUT_CHAR) return;
    if (input_string[0])
    {
        if ((parse_get_numbers(input_string, &input, sizeof(input), &elem, PARSE_INTEGER_SIZE_UINT32, PARSE_RADIX_DEC) != 0) ||
            (elem != 1) || ((input > 0) && (input < 5)) || (input > 65535))
        {
            command_error();
            return;
        }

        system_config.time_auto = input;
        if (input == 0) uart_set_timeout(0);    // Disable timeout immediately if this is desired.
    }
    
    uart_write_string("Poll time (ms): ");
    uart_write_dec_int(system_config.time_poll);
    uart_write_nl();

    uart_write_prompt("New poll time: ");
    uart_read_string(input_string, sizeof(input_string), UART_PROMPT_ECHO);
    if (input_string[0] == UART_TIMEOUT_CHAR) return;
    if (input_string[0])
    {
        if ((parse_get_numbers(input_string, &input, sizeof(input), &elem, PARSE_INTEGER_SIZE_UINT32, PARSE_RADIX_DEC) != 0) ||
            (elem != 1) || (input > 65535))
        {
            command_error();
            return;
        }

        system_config.time_poll = input;
    }

    uart_write_string("Delay time (ms):   ");
    uart_write_dec_int(system_config.time_delay);
    uart_write_nl();
    
    uart_write_prompt("New delay time: ");
    uart_read_string(input_string, sizeof(input_string), UART_PROMPT_ECHO);
    if (input_string[0] == UART_TIMEOUT_CHAR) return;
    if (input_string[0])
    {
        if ((parse_get_numbers(input_string, &input, sizeof(input), &elem, PARSE_INTEGER_SIZE_UINT32, PARSE_RADIX_DEC) != 0) ||
            (elem != 1) || (input > 65535))
        {
            command_error();
            return;
        }

        system_config.time_delay = input;
    }
}

void config_set_cache()
{
    uint8_t input;
    uint32_t input32;
    uint16_t elem;

    uart_write_string("Tag cache size: ");
    uart_write_dec_int(system_config.cache_size);
    uart_write_nl();

    uart_write_prompt("New cache size: ");
    uart_read_string(input_string, sizeof(input_string), UART_PROMPT_ECHO);
    if (input_string[0] == UART_TIMEOUT_CHAR) return;
    if (input_string[0])
    {
        if ((parse_get_numbers(input_string, &input, sizeof(input), &elem, PARSE_INTEGER_SIZE_UINT8, PARSE_RADIX_DEC) != 0) ||
            (elem != 1))
        {
            command_error();
            return;
        }

        if (input > MAX_CACHE_ENTRIES)
        {
            command_error();
            return;
        }
    
        system_config.cache_size = input;
    }
    
    uart_write_string("Cache timeout (s): ");
    uart_write_dec_int(system_config.cache_timeout);
    uart_write_nl();

    uart_write_prompt("New cache timeout: ");
    uart_read_string(input_string, sizeof(input_string), UART_PROMPT_ECHO);
    if (input_string[0] == UART_TIMEOUT_CHAR) return;
    if (input_string[0])
    {
        if ((parse_get_numbers(input_string, &input32, sizeof(input32), &elem, PARSE_INTEGER_SIZE_UINT32, PARSE_RADIX_DEC) != 0) ||
            (elem != 1))
        {
            command_error();
            return;
        }

        system_config.cache_timeout = (uint16_t)input32;
    }
}

void config_set_antenna_switch()
{
    uint8_t input;
    uint16_t elem;
    
    uart_write_string("Start antenna index: ");
    uart_write_dec_int(system_config.ant_start);
    uart_write_nl();

    uart_write_prompt("New start antenna: ");
    uart_read_string(input_string, sizeof(input_string), UART_PROMPT_ECHO);
    if (input_string[0] == UART_TIMEOUT_CHAR) return;
    if (input_string[0])
    {
        if ((parse_get_numbers(input_string, &input, sizeof(input), &elem, PARSE_INTEGER_SIZE_UINT8, PARSE_RADIX_DEC) != 0) ||
            (elem != 1) || (input >= MAX_ANTENNAS))
        {
            command_error();
            return;
        }

        system_config.ant_start = input;
    }
    
    uart_write_string("Stop antenna index: ");
    uart_write_dec_int(system_config.ant_stop);
    uart_write_nl();

    uart_write_prompt("New stop antenna: ");
    uart_read_string(input_string, sizeof(input_string), UART_PROMPT_ECHO);
    if (input_string[0] == UART_TIMEOUT_CHAR) return;
    if (input_string[0])
    {
        if ((parse_get_numbers(input_string, &input, sizeof(input), &elem, PARSE_INTEGER_SIZE_UINT8, PARSE_RADIX_DEC) != 0) ||
            (elem != 1) || (input >= MAX_ANTENNAS))
        {
            command_error();
            return;
        }

        system_config.ant_stop = input;
    }
}

void config_set_battery()
{
    uint32_t input;
    uint16_t elem;
    
    uart_write_string("Battery off threshold (mv): ");
    uart_write_dec_int(system_config.batt_off_threshold);
    uart_write_nl();

    uart_write_prompt("New off threshold (mv): ");
    uart_read_string(input_string, sizeof(input_string), UART_PROMPT_ECHO);
    if (input_string[0] == UART_TIMEOUT_CHAR) return;
    if (input_string[0])
    {
        if ((parse_get_numbers(input_string, &input, sizeof(input), &elem, PARSE_INTEGER_SIZE_UINT32, PARSE_RADIX_DEC) != 0) ||
            (elem != 1) || (input > 65535))
        {
            command_error();
            return;
        }

        system_config.batt_on_threshold = (uint16_t)input;
    }
    
    uart_write_string("Battery on threshold (mv): ");
    uart_write_dec_int(system_config.batt_on_threshold);
    uart_write_nl();

    uart_write_prompt("New on threshold: ");
    uart_read_string(input_string, sizeof(input_string), UART_PROMPT_ECHO);
    if (input_string[0] == UART_TIMEOUT_CHAR) return;
    if (input_string[0])
    {
        if ((parse_get_numbers(input_string, &input, sizeof(input), &elem, PARSE_INTEGER_SIZE_UINT8, PARSE_RADIX_DEC) != 0) ||
            (elem != 1) || (input > 65535))
        {
            command_error();
            return;
        }

        system_config.batt_on_threshold = input;
    }
}



void config_print()
{
    uart_write_string("Date/Time:         ");
    rtc_print_date_time(&rtc_time, 0);
    uart_write_nl();
    
    uart_write_string("Board ID:          \"");
    uart_write_string(system_config.board_id);
    uart_write('"');
    uart_write_nl();

    uart_write_string("Auto start (s):    ");
    uart_write_dec_int(system_config.time_auto);
    uart_write_nl();
    
    uart_write_string("Poll time (ms):    ");
    uart_write_dec_int(system_config.time_poll);
    uart_write_nl();

    uart_write_string("Delay time (ms):   ");
    uart_write_dec_int(system_config.time_delay);
    uart_write_nl();
    
    uart_write_string("Battery off (mV):  ");
    uart_write_dec_int(system_config.batt_off_threshold);
    uart_write_nl();
    
    uart_write_string("Battery on (mV):   ");
    uart_write_dec_int(system_config.batt_on_threshold);
    uart_write_nl();

    uart_write_string("Tag cache size:    ");
    uart_write_dec_int(system_config.cache_size);
    uart_write_nl();
    
    uart_write_string("Cache timeout (s): ");
    uart_write_dec_int(system_config.cache_timeout);
    uart_write_nl();
    
    uart_write_string("Antenna start:     ");
    uart_write_dec_int(system_config.ant_start);
    uart_write_nl();

    uart_write_string("Antenna stop:      ");
    uart_write_dec_int(system_config.ant_stop);
    uart_write_nl();
}

    
