/*============================================================================
 * eeprom.c
 *	Library to read/write 2 Mbit EEPROM part.
 * Author: Jon W. Wallace 
 *==========================================================================*/

#include <inttypes.h>

#include "i2c.h"
#include "eeprom.h"
#include "uart.h"
//#include "rtc_timer.h"
#include "timer_ms.h"

#define ST_2MBIT_ADDR           0xa0

#define WRITE_TIMEOUT_MS        500
#define WRITE_TIMEOUT_TICKS     (MS_TO_TIMER1(WRITE_TIMEOUT_MS))

#define EEPROM_VDD_PORT         (PORTC)
#define EEPROM_VDD_LAT          (LATC)
#define EEPROM_VDD_TRIS         (TRISC)
#define EEPROM_VDD_ANSEL        (ANSELC)
#define EEPROM_VDD_BIT          (1 << 2) // RC2

void eeprom_init()
{
    // Use a digital pin on PIC to power the EEPROM.  This way we can reset it
    // by cycling power to avoid a stuck condition on I2C bus.  Can also power it
    // down if needed.
    EEPROM_VDD_LAT &= ~EEPROM_VDD_BIT;         // Turn power off initially
    EEPROM_VDD_ANSEL &= ~EEPROM_VDD_BIT;        // Turn off analog
    EEPROM_VDD_TRIS &= ~EEPROM_VDD_BIT;         // Set to be output

    // Hold EEPROM OFF to ensure it is reset
    timer_ms_delay(250);

    // Turn eeprom ON
    EEPROM_VDD_LAT |= EEPROM_VDD_BIT;
    timer_ms_delay(250);
}

int8_t eeprom_write_bytes(uint32_t address, uint8_t *data, uint8_t count)
{
    int8_t ccode;
    uint8_t i;
    int8_t do_restart;
    uint32_t address_next;
    uint16_t t1, dt;
    
    do_restart = 1;

    // Send data bytes
    for (i=0; ; )
    {
        // Set address either before first byte written or when the page
        // changes.
        if (do_restart)
        {
            // Try to write
            //t1 = rt_timer_read();
            t1 = timer_ms_read_uint16();
            for ( ; ; )
            {
                // Address chip and send upper address bits
                if (i2c_start(ST_2MBIT_ADDR | (((address >> 16) & 0x3) << 1),
                              I2C_START_WRITE, I2C_ACK) == 0)
                {
                    // Send next two address bytes
                    if ((i2c_write_byte((address >> 8) & 0xff, I2C_ACK) == 0) &&
                        (i2c_write_byte((address & 0xff), I2C_ACK) == 0))
                    {

                        // Success.  Device should now be able to accept write.
                        break;
                    }
                }
                
                // Device not present or not able to accept a write yet.  Check if we have timed out.
                //dt = rt_timer_read()-t1;
                dt = timer_ms_read_uint16()-t1;
                //if (dt > WRITE_TIMEOUT_TICKS)
                if (dt > WRITE_TIMEOUT_MS)                
                {
                    ccode = -1;
                    goto backout_base;
                }

                // Timeout not expired.  Keep trying.
                i2c_stop();
            }
        }
        do_restart = 0;

        if (i >= count) break;
        
        if (i2c_write_byte(data[i], I2C_ACK) < 0)
        {
            ccode = -3;
            goto backout_base;
        }

        if (++i >= count) break;
        
        address_next = address+1;
        if ((address_next & 0x0003ff00) != (address & 0x0003ff00))
        {
            /* New page, so restart next time. */
            do_restart = 1;
            i2c_stop();
        }
        address = address_next;
    }

    ccode=0;
    
 backout_base: 
    i2c_stop();
    return(ccode);
}

int8_t eeprom_read_bytes(uint32_t address, uint8_t *data, uint8_t count)
{
    uint8_t i;
    int8_t ccode;
    
    // Set up address by writing with no data.
    eeprom_write_bytes(address, 0, 0);

    // Address chip
    if (i2c_start(ST_2MBIT_ADDR, I2C_START_READ, I2C_ACK) < 0)
    {
        ccode = -1;
        goto backout_base;
    }

    // Read data
    for (i=0; i<count; i++)
    {
        // Read byte, send NAK on last one.
        if (i2c_read_byte(&data[i], (i < count-1)) < 0)
        {
            ccode = -2;
            goto backout_base;
        }
    }

    ccode=0;

 backout_base:
    i2c_stop();
    return(ccode);
}

void eeprom_sleep()
{
    // Turn eeprom (and BT) OFF
    EEPROM_VDD_LAT &= ~EEPROM_VDD_BIT;
}
