/*============================================================================
 * i2c.c
 *      I2C library.
 * Author: Jon W. Wallace 
 *==========================================================================*/

#include <xc.h>
#include <inttypes.h>
#include "mcc_generated_files/mcc.h"
#include "uart.h"

#include "i2c.h"

// SCL1 = RC3 = pin 14
// SDA1 = RC4 = pin 15

#define I2C_DEBUG               1

#define I2C_PORT_ANSEL          ANSELC
#define I2C_PORT_TRIS           TRISC
#define I2C_PORT                PORTC
#define I2C_PORT_LAT            LATC
#define I2C_PORT_PINS           ((1 << 3) | (1 << 4))

#define SSPxCLKPPS      SSP1CLKPPS      // I2C CLK1 (Input)
#define SSPxCLKPPS_VAL  0x13            //RC3
#define SPPxDATPPS      SSP1DATPPS      // I2C DAT1 (Input)
#define SSPxDATPPS_VAL  0x14            //RC4

#define CLK_RxxPPS      RC3PPS  // RC3 (Output)
#define CLK_RxxPPS_VAL  0x13    // SCL1
#define DAT_RxxPPS      RC4PPS  // RC4 (Output)
#define DAT_RxxPPS_VAL  0x14    // SDA1

#define SSPxCON1                SSP1CON1
#define SSPxCON1_SSPM_I2C_MASTER        0x08
#define SSPxCON1_CKP                    (1 << 4)        // Enable clock
#define SSPxCON1_SSPEN                  (1 << 5)        // Enable serial port

#define SSPxCON2                SSP1CON2
#define SSPxCON2_SEN            (1 << 0)
#define SSPxCON2_RSEN           (1 << 1)
#define SSPxCON2_PEN            (1 << 2)
#define SSPxCON2_RCEN           (1 << 3)
#define SSPxCON2_ACKEN          (1 << 4)
#define SSPxCON2_ACKDT          (1 << 5)
#define SSPxCON2_ACKSTAT        (1 << 6)
#define SSPxCON2_GCEN           (1 << 7)

#define SSPxADD                 SSP1ADD

#define SSPxIF_BIT              (1 << 0)
#define SSPxIF_REG              PIR3
#define SSPxBUF                 SSP1BUF
#define SSPxBUF_R               (1 << 0)  // R!W bit

#define SSPxSTAT                SSP1STAT
#define SSPxSTAT_RW             (1 << 2)

#define PCLK            (_XTAL_FREQ)
#define I2C_BAUD        (100000L)
#define I2C_BAUD_BRG    (PCLK/I2C_BAUD/4)-1

// Macro to wait for interrupt complete and clear it.
#define I2C_WAIT_IDLE  \
    while ((SSPxSTAT & SSPxSTAT_RW) || (SSPxCON2 & 0x1f)) 

void i2c_init(void)
{
    // Turn off analog function and make inputs on pins.
    I2C_PORT_ANSEL &= ~(I2C_PORT_PINS);
    I2C_PORT_TRIS |= (I2C_PORT_PINS);

    /*
    uart_write_string("I2C mapping at boot:\r\n");
    uart_write_hex_byte(SSPxCLKPPS);
    uart_write_hex_byte(SPPxDATPPS);
    uart_write_hex_byte(CLK_RxxPPS);
    uart_write_hex_byte(DAT_RxxPPS);
    uart_write_nl();
    */
    
    // Map bidirectional I2C peripheral pins to digital pins
    SSPxCLKPPS = SSPxCLKPPS_VAL;
    SPPxDATPPS = SSPxDATPPS_VAL;
    CLK_RxxPPS = CLK_RxxPPS_VAL;
    DAT_RxxPPS = DAT_RxxPPS_VAL;
    
    // Set control register.  Set for master mode and enable it.
    SSPxCON1 = SSPxCON1_SSPM_I2C_MASTER | SSPxCON1_SSPEN;
    SSPxCON2 = 0;

    // Set baud rate
    SSPxADD = I2C_BAUD_BRG;

    //SSPxSTAT = 0xc0;
}


int8_t i2c_start(uint8_t addr, uint8_t r_not_w, uint8_t check_ack)
{
    // Generate a start bit
    I2C_WAIT_IDLE;
    SSPxCON2 |= SSPxCON2_SEN;

    // Load the slave address
    I2C_WAIT_IDLE;
    if (r_not_w)
    {
        SSPxBUF = addr | SSPxBUF_R;
    }
    else
    {
        SSPxBUF = addr & ~SSPxBUF_R;
    }

    // Check if ACK was received.
    if (check_ack && (SSPxCON2 & SSPxCON2_ACKSTAT)) return(-1);

    return(0);
}

int8_t i2c_write_byte(uint8_t data, uint8_t check_ack)
{
    I2C_WAIT_IDLE;
    SSPxBUF = data;

    // Check if ACK was received.
    if (check_ack && (SSPxCON2 & SSPxCON2_ACKSTAT)) return(-1);
    
    return(0);
}

int8_t i2c_read_byte(uint8_t *data, uint8_t send_ack)
{
    uint8_t i;
    int8_t ccode;

    // Set up flag for the ACK
    if (send_ack)
    {
        SSPxCON2 &= ~SSPxCON2_ACKDT;
    }
    else
    {
        SSPxCON2 |= SSPxCON2_ACKDT;
    }
    
    // Set to receive data
    I2C_WAIT_IDLE;
    SSPxCON2 |= SSPxCON2_RCEN;
    I2C_WAIT_IDLE;

    // Get the data
    *data = SSPxBUF;
    //I2C_WAIT_IDLE;

    // Acknowledge
    SSPxCON2 |= SSPxCON2_ACKEN;

    // Success
    return(0);
}

int8_t i2c_stop(void)
{
    // Send stop condition
    I2C_WAIT_IDLE;
    SSPxCON2 |= SSPxCON2_PEN;
    return(0);
}

void i2c_sleep()
{
    // Turn off I2C module
    SSPxCON1 = 0;

    // Set assignments to be power on defaults.
    SSPxCLKPPS = 0x13;
    SPPxDATPPS = 0x14;
    CLK_RxxPPS = 0x00;
    DAT_RxxPPS = 0x00;
    
    // Set CLK/DATA pins to outputs and drive high
    // Uses least power because we have external pullups.
    I2C_PORT_TRIS &= (~I2C_PORT_PINS);
    I2C_PORT_LAT |= I2C_PORT_PINS;
}
