/*============================================================================
 * main.c
 *      Main entry point into RFID reader firmware.
 *
 * Author: Jon Wallace, Lafayette College
 *
 *==========================================================================*/

#define BOOT_DEBUG_MESSAGE
//#define SLEEP_USE_32KHZ_CLOCK

#define VB0GPR_SLEEP_CODE         0xa5

#include <xc.h>

#include "mcc_generated_files/mcc.h"

#include "rtc_timer.h"
#include "timer.h"
#include "timer_ms.h"
#include "uart.h"
#include "command.h"
#include "i2c.h"
#include "eeprom.h"
#include "switch.h"
#include "rfid.h"
#include "config.h"
#include "adc.h"
#include "test.h"
#include "tag_log.h"
#include "battery.h"

void timer_test_ms(void);

//#define OSCEN_SOSCEN (1 << 3)
//#define OSCEN_ADOEN  (1 << 2)

#define CPUDOZE_IDLEN (1 << 7)

void do_sleep()
{
    GINT_OFF;
    
    // Put modules to sleep as necessary
    //adc_sleep();
    rfid_sleep();
    switch_sleep();
    i2c_sleep();
    uart_sleep();
    eeprom_sleep();
    timer_sleep();
    timer_ms_sleep();
    battery_sleep();

#ifdef SLEEP_USE_32KHZ_CLOCK    
    // Switch the processor to now use the 32kHz SOSC oscillator
    OSCCON1 = 0x80;

    // Wait until the switch is complete
    while (!(OSCCON3 & 0x10));

    // Switch is complete, only keep needed low power oscillators
    OSCEN = OSCEN_SOSCEN | OSCEN_ADOEN;
#endif

    // Change operation of RT timer to allow wakeup on overflow
    rt_sleep();
    
    for ( ; ; )
    {
    
        // Execute sleep instruction
        CPUDOZE &= ~CPUDOZE_IDLEN;
        SLEEP();

        // Reach here on TMR1 overflow (about 16 seconds)
        rt_init();
        //uart_init();
        //uart_write_string("Woke up.  Checking battery...\r\n");
        
        if (battery_sleep_wakeup_ok())
        {
            break;
        }

        rt_sleep();
        uart_sleep();
    }
    
    // Set a value in the GP persistent register so we know it
    // was a wakeup.
    VB0GPR = VB0GPR_SLEEP_CODE;

    RESET();
}

void main(void)
{
    // initialize the device.  Uses MCC generated files.
    SYSTEM_Initialize();

    /*
    TRISA &= ~(1 << 4);
    ANSELA &= ~(1 << 4);
    LATA |= (1 << 4);
    for ( ; ; );
    */

    // Initalize modules
    rt_init();
    uart_init();
    rtc_init();
    adc_init();
    timer_init();
    timer_ms_init();
    i2c_init();
    switch_init();
    config_init();
    rfid_init();
    command_init();
    test_init();

    uart_write_string("Boot\r\n");
    
    rt_delay(500);

#ifdef BOOT_DEBUG_MESSAGE    
    uart_write_string("G int on\r\n");
#endif    
    // Enable the Global Interrupts
    //INTERRUPT_GlobalInterruptEnable();
    INTCON |= 0x80;

#ifdef BOOT_DEBUG_MESSAGE    
    uart_write_string("P int on\r\n");
#endif
    
    // Enable the Peripheral Interrupts
    //INTERRUPT_PeripheralInterruptEnable();
    INTCON |= 0x40;
    
    // Disable the Global Interrupts
    //INTERRUPT_GlobalInterruptDisable();

    // Disable the Peripheral Interrupts
    //INTERRUPT_PeripheralInterruptDisable();

    // Delay to ensure that the oscillators are stable
    timer_ms_delay(1000);

    // Initialize devices that depend on timing to start correctly
    eeprom_init();

    // Load configuration
#ifdef BOOT_DEBUG_MESSAGE    
    uart_write_string("Config load\r\n");
#endif    
    switch (config_load())
    {
        case -1:
            uart_write_string("Config read failed! Using default.\n");
            config_default();
            break;
            
        case -2:
            uart_write_string("Config checksum error! Using default.\n");
            config_default();
            break;
    }

#ifdef BOOT_DEBUG_MESSAGE    
    uart_write_string("Read log index\r\n");
#endif    
    tag_log_init();

    battery_check();
    
#ifdef BOOT_DEBUG_MESSAGE    
    uart_write_string("Start shell\r\n");
#endif

    /*
    uart_write_string("VB0GPR=");
    uart_write_hex_byte(VB0GPR);
    uart_write_nl();
    */
    
    if (VB0GPR == VB0GPR_SLEEP_CODE)
    {
        tag_log_write_code(TAG_LOG_CODE_WAKE);
    }
    else
    {
        tag_log_write_code(TAG_LOG_CODE_POWER);
    }
    VB0GPR = 0;
    
    command_run();
}

void __interrupt () master_isr(void)
{
    timer_ms_isr();
}
