/*============================================================================
 * parse.c
 *      Functions to parse input.
 * Author: Jon W. Wallace 
 *==========================================================================*/

#include "parse.h"

#define PARSE_INTEGER_SIZE_UINT8                1
#define PARSE_INTEGER_SIZE_UINT32               4

#define PARSE_RADIX_HEX                         1
#define PARSE_RADIX_DEC                         2

/*----------------------------------------------------------------------------
 * parse_hex_char()
 *      Convert a hex character to binary number 0-15.
 *--------------------------------------------------------------------------*/
int8_t parse_hex_char(parse_char c)
{
    if ((c >= '0') && (c <= '9'))
    {
        return(c - '0');
    }
    else if ((c >= 'a') && (c <= 'f'))
    {
        return(c - 'a' + 10);
    }
    else if ((c >= 'A') && (c <= 'F'))
    {
        return(c - 'A' + 10);
    }
    else
    {
        return(-1);
    }
}

/*----------------------------------------------------------------------------
 * parse_dec_char()
 *      Convert a decimal character to binary number 0-9.
 *--------------------------------------------------------------------------*/
int8_t parse_dec_char(parse_char c)
{
    if ((c >= '0') && (c <= '9'))
    {
        return(c - '0');
    }
    else
    {
        return(-1);
    }
}

/*----------------------------------------------------------------------------
 * parse_get_numbers
 *      Reads 32-bit hex numbers (separated by spaces)
 *--------------------------------------------------------------------------*/
int8_t parse_get_numbers(parse_char *buffer,
                         void *buffer_out,
                         uint16_t buffer_out_size,
                         uint16_t *elements_returned,
                         uint8_t int_size,
                         uint8_t radix)
{
    uint16_t max_elements;
    
    uint16_t num_idx;
    uint32_t num1;

    uint8_t  *p_u8;
    uint32_t *p_u32;

    int8_t digit;
    
    // Get maximum number of elements of and pointer to output buffer.
    switch(int_size)
    {
        case PARSE_INTEGER_SIZE_UINT8:
            max_elements = buffer_out_size;
            p_u8 = (uint8_t *) buffer_out;
            break;

        case PARSE_INTEGER_SIZE_UINT32:
            max_elements = buffer_out_size/sizeof(uint32_t);
            p_u32 = (uint32_t *) buffer_out;
            break;

        default:
            return(-1);
    }
    
    
    num_idx = 0;

    // Forever
    for ( ; ; )
    {
        // Make sure we have storage left
        if (num_idx >= max_elements) break;
        
        // Go until we find a non-space
        while ((*buffer==' ') || (*buffer=='\n') || (*buffer=='\r') || (*buffer==',') || (*buffer==':') || (*buffer == '/'))
        {
            buffer++;
        }
        if (*buffer == 0) break;
        
        // Read as many digits as possible
        num1 = 0;
        switch(radix)
        {
            case PARSE_RADIX_DEC:
                while ((digit = parse_dec_char(*buffer)) >= 0)
                {
                    num1 = num1*10 + (uint32_t)digit;
                    buffer++;
                }
                break;
            
            case PARSE_RADIX_HEX:
                while ((digit = parse_hex_char(*buffer)) >= 0)
                {
                    num1 = (num1 << 4) + (uint32_t)digit;
                    buffer++;
                }
                break;
            default:
                return(-2);
        }

        // Store value
        switch(int_size)
        {
            case PARSE_INTEGER_SIZE_UINT8:
                p_u8[num_idx] = (uint8_t) num1;
                break;

            case PARSE_INTEGER_SIZE_UINT32:
                p_u32[num_idx] = (uint32_t) num1;
                break;
        }
        num_idx++;
    }

    *elements_returned = num_idx;
    return(0);
}

