/*============================================================================
 * rfid.c
 *      Code to communicate with the U2270B RFID module.  Uses a
 *      manchester coding strategy.  We use TIMER0 to allow us to time the
 *      pulses received.
 * Author: Jon W. Wallace 
 *==========================================================================*/

#include <xc.h>
#include <inttypes.h>

#include "global.h"
#include "rtc_timer.h"
#include "timer.h"
#include "timer_ms.h"
#include "rfid.h"
#include "uart.h"

// Enable a debug GPIO pin (DBG pin)
// Needed for other GPIO debug modes to work below
// #define DEBUG_OUTPUT
// Copy the sampled input to the output to make sure no noise 
// #define DEBUG_MODE_MIRROR_INPUT
// Put markers at critical bit timing events to allow tuning
// #define DEBUG_MODE_BIT_TIMING

//#define DEBUG_MESSAGES                

#define TS1_TICS          TIMER_US_TO_TICKS(180)        // Short pulse minimum
#define TS2_TICS          TIMER_US_TO_TICKS(360)        // Short pulse length max
#define TL1_TICS          TIMER_US_TO_TICKS(420)        // Long pulse minimum
#define TL2_TICS          TIMER_US_TO_TICKS(600)        // Long pulse maximum

#define EDGE_TIMEOUT      TIMER_US_TO_TICKS(800)       // Timeout when looking for initial edge

// Time from beginning of bit period to where we sample bit.  Should be
// 3/4 of a bit time (380us).  Find that this puts us too close to the
// edge in practice.  Value was tuned using a scope.
#define TWAIT_TICS        TIMER_US_TO_TICKS(340)        

#define RFID_STANDBY1_PORT       (PORTB)
#define RFID_STANDBY1_LAT        (LATB)
#define RFID_STANDBY1_TRIS       (TRISB)
#define RFID_STANDBY1_ANSEL      (ANSELB)
#define RFID_STANDBY1_BIT        (1 << 4)        // RB4 = Pin 25

#define RFID_STANDBY2_PORT       (PORTA)
#define RFID_STANDBY2_LAT        (LATA)
#define RFID_STANDBY2_TRIS       (TRISA)
#define RFID_STANDBY2_ANSEL      (ANSELA)
#define RFID_STANDBY2_BIT        (1 << 7)        // RA7

volatile uint8_t *stdby_port[MAX_PORTS] = 
    {&RFID_STANDBY1_PORT, &RFID_STANDBY1_PORT, &RFID_STANDBY1_PORT, &RFID_STANDBY1_PORT, &RFID_STANDBY1_PORT, &RFID_STANDBY1_PORT, &RFID_STANDBY1_PORT, &RFID_STANDBY1_PORT,
     &RFID_STANDBY2_PORT, &RFID_STANDBY2_PORT, &RFID_STANDBY2_PORT, &RFID_STANDBY2_PORT};

volatile uint8_t *stdby_lat[MAX_PORTS] = 
    {&RFID_STANDBY1_LAT, &RFID_STANDBY1_LAT, &RFID_STANDBY1_LAT, &RFID_STANDBY1_LAT, &RFID_STANDBY1_LAT, &RFID_STANDBY1_LAT, &RFID_STANDBY1_LAT, &RFID_STANDBY1_LAT,
     &RFID_STANDBY2_LAT, &RFID_STANDBY2_LAT, &RFID_STANDBY2_LAT, &RFID_STANDBY2_LAT};

uint8_t stdby_bit[MAX_PORTS] =
    {RFID_STANDBY1_BIT, RFID_STANDBY1_BIT, RFID_STANDBY1_BIT, RFID_STANDBY1_BIT, RFID_STANDBY1_BIT, RFID_STANDBY1_BIT, RFID_STANDBY1_BIT, RFID_STANDBY1_BIT,
     RFID_STANDBY2_BIT, RFID_STANDBY2_BIT, RFID_STANDBY2_BIT, RFID_STANDBY2_BIT};

#define RFID_OUTPUT_PORT        (PORTB)
#define RFID_OUTPUT_TRIS        (TRISB)
#define RFID_OUTPUT_ANSEL       (ANSELB)
#define RFID_OUTPUT_BIT         (1 << 5)        // RB5 = Pin 26

#ifdef DEBUG_OUTPUT
#define RFID_DEBUG_PORT         (PORTA)
#define RFID_DEBUG_LAT          (LATA)
#define RFID_DEBUG_TRIS         (TRISA)
#define RFID_DEBUG_ANSEL        (ANSELA)
#define RFID_DEBUG_BIT_POS      7
#define RFID_DEBUG_BIT          (1 << (RFID_DEBUG_BIT_POS))  
#define RFID_DEBUG_SET(v)    RFID_DEBUG_LAT = ((RFID_DEBUG_LAT & ~RFID_DEBUG_BIT) | (v << RFID_DEBUG_BIT_POS))
#define RFID_DEBUG_TOGGLE       (RFID_DEBUG_LAT ^= RFID_DEBUG_BIT)
#endif

// RFID is usually a macro, but in the mirror input (debug) mode, it is a
// function that outputs all read values to the debug pin.
#ifndef DEBUG_MODE_MIRROR_INPUT
#define RFID_READ               ((RFID_OUTPUT_PORT & RFID_OUTPUT_BIT) != 0)
#else
inline uint8_t do_rfid_read() 
{
    static uint8_t rfid_debug_tmp;
    rfid_debug_tmp=((RFID_OUTPUT_PORT & RFID_OUTPUT_BIT) != 0);
    RFID_DEBUG_SET(rfid_debug_tmp);
    return(rfid_debug_tmp);
}
#define RFID_READ       do_rfid_read()
#endif

#define GET_PULSE_SHORT          0
#define GET_PULSE_LONG           1
#define GET_PULSE_INVALID       -2

#define GET_PARITY5(b)   (((b)&0x1) ^ (((b)>>1)&0x1) ^ (((b)>>2)&0x1) ^ (((b)>>3)&0x1) ^ (((b)>>4)&0x1))
    
void rfid_init()
{
    // Set up the lines we use to communicate with the board.
    RFID_STANDBY1_LAT &= ~RFID_STANDBY1_BIT;
    RFID_STANDBY1_ANSEL &= ~RFID_STANDBY1_BIT;
    RFID_STANDBY1_TRIS &= ~RFID_STANDBY1_BIT;

    RFID_STANDBY2_LAT &= ~RFID_STANDBY2_BIT;
    RFID_STANDBY2_ANSEL &= ~RFID_STANDBY2_BIT;
    RFID_STANDBY2_TRIS &= ~RFID_STANDBY2_BIT;
    
    RFID_OUTPUT_ANSEL &= ~RFID_OUTPUT_BIT;
    RFID_OUTPUT_TRIS |= RFID_OUTPUT_BIT;  // Input

    // Setup debug port to allow timing to be checked
#ifdef DEBUG_OUTPUT
    RFID_DEBUG_LAT &= ~RFID_DEBUG_BIT;
    RFID_DEBUG_ANSEL &= ~RFID_DEBUG_BIT;
    RFID_DEBUG_TRIS &= ~RFID_DEBUG_BIT;
#endif
}

// State variables shared by functions
static uint32_t timeout_ms_t1, timeout_ms_t2;

// Global timeout
// static uint16_t timeout_rt_ticks;
static uint32_t timeout_ms;

uint8_t check_timeout()
{
    uint32_t dt;
    
    // Check for timeout.
    timeout_ms_t2 = timer_ms_read();           

    dt = timeout_ms_t2 - timeout_ms_t1;
    if (dt > timeout_ms)
    {
        return(1);
    }
    return(0);
}

// Beginning time reference point for all interval computations.  May be a transition, or
// the transition + TWAIT.  More accurate to do it this way than to have each function
// read time at the beginning, hoping that not much time has passed.
static timer_t last_event_time = 0;             

// Waits for a transition on OUT or a timeout.  Returns the bit state
// after the transition (0 or 1) or -1 for a timeout.
int8_t wait_trans_timeout(timer_t timeout)
{
    uint8_t i;
    timer_t t2, dt;
    uint8_t out, out_D;

    // Store current value of output
    out_D = RFID_READ;
    
    // Wait for a transition
    for ( ; ; )
    {
        // Check for a transition
        t2 = TIMER_READ;
#ifdef RFID_READ_FILTER
        out = rfid_read_filter();
#else
        out = RFID_READ;
#endif
        if (out != out_D) break;   // Got transition.  OK.


        // Compute time that has passed.  Note that for wrapping to work correctly,
        // need to assign to another unsigned variable (dt in this case).
        dt = t2-last_event_time;  
        if (dt >= timeout)
        {
            last_event_time = t2;
            return(-1);
        }

    }
    last_event_time = t2;
    return((int8_t)out);
}

// Waits the interval TWAIT and then updates the event time.
void do_twait()
{
    timer_t t2, dt;

    do
    {
        t2 = TIMER_READ;
        dt = t2-last_event_time;
    }
    while (dt < TWAIT_TICS);

    last_event_time = t2;
}
        
// Goes until we find a wide pulse or time out.
int8_t get_sync()
{
    timer_t t1, t2, dt;

#ifdef DEBUG_MODE_BIT_TIMING
    RFID_DEBUG_TOGGLE;
    RFID_DEBUG_TOGGLE;
    RFID_DEBUG_TOGGLE;
    RFID_DEBUG_TOGGLE;
#endif
    
    // Find a first edge
    if (wait_trans_timeout(EDGE_TIMEOUT) < 0) return(-1);
    t1 = last_event_time;
    
    for ( ; ; )
    {
        if (wait_trans_timeout(EDGE_TIMEOUT) < 0) return(-1);
        dt = last_event_time-t1;
        t1 = last_event_time;
        
        // Check what type of pulse this is.
        if (dt < TS1_TICS)
        {
            // Runt pulse.  Error.
            return(-1);
        }
        else if (dt < TS2_TICS)
        {
            // Short pulse.  Not what we want but OK.   Keep trying.
            continue;
        }
        else if (dt < TL1_TICS)
        {
            // Between long and short.  Error.
            return(-1);
        }
        else if (dt < TL2_TICS)
        {
            // Long pulse.  What we want.
            break;
        }
    }

    // Once the long pulse is found, wait 3/4 of a bit time (TWAIT) which brings us
    // just before the middle of a bit period.
    do_twait();
    return(0);
}

inline int8_t get_bit()
{
    int8_t bit;

#ifdef DEBUG_MODE_BIT_TIMING
    RFID_DEBUG_TOGGLE;
    RFID_DEBUG_TOGGLE;
#endif
    
    // Wait for an edge.  Must occur within short pulse time.
    // Turn off interrupts to ensure we don't miss it due to
    // an interrupt firing.
    GINT_OFF;
    if ((bit=wait_trans_timeout(TS2_TICS)) < 0)
    {
        GINT_ON;
        return(-1);
    }
    GINT_ON;
    
    do_twait();

    // Bit value is value before the transition
    return(!bit);
}

uint8_t bit_data[11];
uint8_t rfid_data_out[5];

int8_t check_data()
{
    uint8_t i;
    uint8_t p;
    
    // Check parity on rows
    for (i=0; i<10; i++)
    {
        if (GET_PARITY5(bit_data[i]) != 0)
        {
            return(-1);
        }
    }

    // Check parity on columns
    p = 0;
    for (i=0; i<11; i++)
    {
        p ^= (bit_data[i] >> 1);
    }
    if (p != 0)
    {
        return(-2);
    }

    return(0);
}

void decode_data()
{
    uint8_t i;
    for (i=0; i<5; i++)
    {
        rfid_data_out[i] = ((bit_data[i*2] >> 1) << 4) | (bit_data[i*2+1] >> 1);
    }
}

uint16_t retry_preamble, retry_data, retry_parity;
int8_t rfid_read(uint32_t timeout_ms_in, uint8_t ant_idx)
{
    int8_t bit;
    uint8_t idx, bit_idx, data1;
    uint8_t ccode;
    uint8_t found0;
    
    retry_preamble = 0;
    retry_data = 0;
    retry_parity = 0;

    // Compute timeout value in RT clock ticks
    //timeout_rt_ticks = MS_TO_TIMER1(timeout_ms);
    timeout_ms = timeout_ms_in;
    
    // Zero out the bit data
    for (idx=0; idx<11; idx++)
    {
        bit_data[idx] = 0;
    }

    // Start timer for checking pulse widths
    timer_reset();
    timer_on();

    // Wake up the RFID reader
    *stdby_lat[ant_idx] |= stdby_bit[ant_idx];

    // Set initial state
    timeout_ms_t1 = timer_ms_read();
    last_event_time = timer_read();
    
 get_sync_edge:
    if (check_timeout())
    {
        ccode = RFID_READ_TIMEOUT;
        goto backout_base;
    }
    
    // Synchronize to regular sample point
    if (get_sync() != 0) goto get_sync_edge;

 get_preamble:

    if (check_timeout())
    {
        ccode = RFID_READ_TIMEOUT;
        goto backout_base;
    }

    // Find a string of 9 1's
    for (idx=0; idx < 9; )
    {
        if ((bit = get_bit()) < 0)
        {
            retry_preamble++;
            goto get_sync_edge;
        }

        if (bit != 1)
        {
            // Not part of preamble.  Retry.
            goto get_preamble;
        }
        else
        {
            idx++;
        }
    }
    
 get_data:

    for (idx=0; idx < 11; idx++)
    {
        data1 = 0;
        for (bit_idx=0; bit_idx < 5; bit_idx++)
        {
            if ((bit = get_bit()) < 0)
            {
                retry_data++;
                goto get_sync_edge;
            }

            data1 = (data1 << 1) | (uint8_t) bit;
        }
        bit_data[idx] = data1;
    }

    if (check_data() != 0)
    {
        ccode = RFID_READ_PARITY;
        retry_parity++;
        goto get_sync_edge;
    }

    decode_data();
    ccode = RFID_READ_OK;
    
 backout_base:
    // Turn off timer (not needed)
    timer_off();

    // Put RFID in standby
    *stdby_lat[ant_idx] &= ~stdby_bit[ant_idx];
    

#ifdef DEBUG_MESSAGES
    rfid_print_stats();
#endif    
    
    return(ccode);
}

void rfid_print_stats()
{
    uart_write_string("Retries (Preable, Data, Parity): ");
    uart_write_dec_int(retry_preamble);
    uart_write(' ');
    uart_write_dec_int(retry_data);
    uart_write(' ');
    uart_write_dec_int(retry_parity);
    uart_write('\r');
    uart_write('\n');
}

void rfid_print_raw_data(void)
{
    uint8_t i;
    uint8_t par;

    par = 0;
    for (i=0; i<11; i++)
    {
        uart_write_hex_byte(bit_data[i] >> 1);
        uart_write(' ');
        uart_write_hex_byte(bit_data[i] & 0x1);
        if (i < 10)
        {
            uart_write(' ');
            uart_write('|');
            uart_write(' ');
            uart_write_hex_byte(GET_PARITY5(bit_data[i]));
        }
        uart_write('\r');
        uart_write('\n');
        par ^= (bit_data[i] >> 1);
    }
    uart_write_string("--\r\n");
    uart_write_hex_byte(par);
    uart_write('\r');
    uart_write('\n');
}

void rfid_print_data(void)
{
    uint8_t i;
    for (i=0; i<5; i++)
    {
        uart_write_hex_byte(rfid_data_out[i]);
    }
    //uart_write('\r');
    //uart_write('\n');
}

void rfid_sleep(void)
{
    // Make sure RFID chip is in standby mode
    RFID_STANDBY1_LAT &= ~RFID_STANDBY1_BIT;
    RFID_STANDBY2_LAT &= ~RFID_STANDBY2_BIT;
}

void rfid_write_standby(uint8_t d)
{
    switch(d)
    {
        case 0:
            RFID_STANDBY1_LAT &= ~RFID_STANDBY1_BIT;
            RFID_STANDBY2_LAT &= ~RFID_STANDBY2_BIT;
            break;
        case 1:
            RFID_STANDBY1_LAT |= RFID_STANDBY1_BIT;
            RFID_STANDBY2_LAT &= ~RFID_STANDBY2_BIT;
            break;
        case 2:
            RFID_STANDBY1_LAT &= ~RFID_STANDBY1_BIT;
            RFID_STANDBY2_LAT |= RFID_STANDBY2_BIT;
            break;
    }
}

uint8_t rfid_read_output()
{
    uint8_t val;

    val = RFID_READ;
    return (val);
}


