/*============================================================================
 * rtc_timer.c
 *      Code to support real-time clock and timer functions.
 * Author: Jon W. Wallace 
 *==========================================================================*/

#include <xc.h>
#include <inttypes.h>
#include "rtc_timer.h"
#include "parse.h"
#include "uart.h"

// Define to read the 16-bit value of TIMER1
#define RT_TIMER1  ((TMR1H << 8) | TMR1L)

#define RTCCON_RTCEN            (1 << 7)
#define RTCCON_RTCWREN          (1 << 5)
#define RTCCON_RTCSYNC          (1 << 4)
#define RTCCON_RTCCLKSEL_SOSC   (0x00)
#define RTCCON_RTCCLKSEL_MSK    (0x03)

#define SLEEP_WAKEUP_INTERVAL_5_SEC     20480

#define PIE4_TMR1IE     (1 << 0)

rtc_time_def rtc_time;

parse_char rtc_input_string[16];
uint8_t rtc_numbers[3];

void rt_init(void)
{
    // Turn on secondary oscillator (32kHz))
    OSCEN |= OSCEN_SOSCEN;

    // Set Timer 1 clock source to be 32kHz oscillator
    T1CLK = T1CLK_SOSC;

    // Clear interrupt flag and source
    PIR4 &= ~PIR4_TMR1IF;
    PIE4 &= ~PIE4_TMR1IE;
    
    // Turn on Timer 1
    T1CON = T1CON_PRESCALE_1 | T1CON_READ_8BIT | T1CON_ON;
}

// rt_sleep()
//      In sleep mode, this oscillator stays on.  However,
//      we change the timer prescaler to be much longer.
//      The timer interrupt is enabled and used to
//      wake up from sleep mode.
void rt_sleep(void)
{
    // Disable the timer
    T1CON = 0;

    TMR1H = 0;
    TMR1L = 0;

    // Set the wake-up interval
    //SMT1PRH = 80; //(uint8_t)(SLEEP_WAKEUP_INTERVAL_5_SEC >> 8);
    //SMT1PRL = 0;  //(uint8_t)(SLEEP_WAKEUP_INTERVAL_5_SEC & 0xff);

    // Enable this as an interrupt source
    PIE4 |= PIE4_TMR1IE;
    // Clear any existing interrupt condition
    PIR4 &= ~PIR4_TMR1IF;
    
    // Turn on the timer
    T1CON = T1CON_PRESCALE_8 | T1CON_READ_8BIT | T1CON_ASYNC;
    T1CON |= T1CON_ON;
}

void rt_sleep_off()
{
    PIE4 &= ~PIE4_TMR1IE;
    T1CON = 0;
}

uint16_t rt_timer_read(void)
{
    return(RT_TIMER1);
}

uint16_t rt_timer_to_ms(uint16_t t1, uint16_t t2)
{
    return((uint16_t)TIMER1_TO_MS(t2-t1));
}

void rt_delay(uint16_t ms)
{
    uint16_t t1, t2;

    t1 = rt_timer_read();

    while (1)
    {
        t2 = rt_timer_read();
        if (rt_timer_to_ms(t1, t2) >= ms) break;
    }
}

uint8_t bcd_to_uint8(uint8_t bcd)
{
    return((bcd >> 4)*10 + (bcd & 0xf));
}

uint8_t uint8_to_bcd(uint8_t i)
{
    return((i % 10) | ((i/10) << 4));
}

void rtc_init()
{
    // Turn on RTCC and make writeable
    RTCCON |= RTCCON_RTCWREN;
    RTCCON |= RTCCON_RTCEN;

    // Make sure correct oscillator is enabled.
    if ((RTCCON & RTCCON_RTCCLKSEL_MSK) != RTCCON_RTCCLKSEL_SOSC)
    {
        RTCCON &= RTCCON_RTCCLKSEL_MSK;
        RTCCON |= RTCCON_RTCCLKSEL_SOSC;
    }
}

void rtc_set_date()
{
    uint8_t year1, month1, day1;

    year1 = uint8_to_bcd(rtc_time.year);
    month1 = uint8_to_bcd(rtc_time.month);
    day1 = uint8_to_bcd(rtc_time.day);

    /*
    uart_write_string("Writing date: ");
    uart_write_hex_byte(month1);
    uart_write('/');
    uart_write_hex_byte(day1);
    uart_write('/');
    uart_write_hex_byte(year1);
    uart_write_nl();
    */
    
    // Wait for it to be OK to write
    while(RTCCON & RTCCON_RTCSYNC);
        
    YEAR = year1;
    MONTH = month1;
    DAY = day1;
}

void rtc_set_time()
{
    uint8_t hour1, min1, sec1;

    hour1 = uint8_to_bcd(rtc_time.hour);
    min1 = uint8_to_bcd(rtc_time.min);
    sec1 = uint8_to_bcd(rtc_time.sec);

    /*
    uart_write_string("Writing time: ");
    uart_write_hex_byte(hour1);
    uart_write(':');
    uart_write_hex_byte(min1);
    uart_write(':');
    uart_write_hex_byte(sec1);
    uart_write_nl();
    */
    
    // Wait for it to be OK to write
    while(RTCCON & RTCCON_RTCSYNC);
        
    HOURS = hour1;
    MINUTES = min1;
    SECONDS = sec1;
}


void rtc_get()
{
    uint8_t year1, month1, day1, hour1, min1, sec1;
    
    // Wait for it to be OK to read
    while(RTCCON & RTCCON_RTCSYNC);

    year1 = YEAR;
    month1 = MONTH;
    day1 = DAY;
    hour1 = HOURS;
    min1 = MINUTES;
    sec1 = SECONDS;
    
    rtc_time.year= bcd_to_uint8(year1);    
    rtc_time.month=bcd_to_uint8(month1);   
    rtc_time.day=  bcd_to_uint8(day1);     
    rtc_time.hour= bcd_to_uint8(hour1);   
    rtc_time.min=  bcd_to_uint8(min1); 
    rtc_time.sec=  bcd_to_uint8(sec1);
    
}

void rtc_user_set_date()
{
    uint16_t elem;
    
    uart_write_prompt("New date (MM/DD/YY): ");
    uart_read_string(rtc_input_string, sizeof(rtc_input_string), UART_PROMPT_ECHO);

    if ((parse_get_numbers(rtc_input_string,
                          rtc_numbers,
                          sizeof(rtc_numbers),
                          &elem,
                          PARSE_INTEGER_SIZE_UINT8,
                          PARSE_RADIX_DEC) != 0) ||
        (elem != 3) ||
        (rtc_numbers[0] > 12) ||
        (rtc_numbers[1] > 31) ||
        (rtc_numbers[2] > 99))
    {
        uart_write_string("Invalid date.\n");
        return;
    }

    rtc_time.year = rtc_numbers[2];
    rtc_time.day = rtc_numbers[1];
    rtc_time.month = rtc_numbers[0];

    rtc_set_date();
}

void rtc_user_set_time()
{
    uint16_t elem;

    uart_write_prompt("New time (HH:MM:SS): ");
    uart_read_string(rtc_input_string, sizeof(rtc_input_string), UART_PROMPT_ECHO);

    if ((parse_get_numbers(rtc_input_string,
                          rtc_numbers,
                          sizeof(rtc_numbers),
                          &elem,
                          PARSE_INTEGER_SIZE_UINT8,
                           PARSE_RADIX_DEC) != 0) || (elem != 3) ||
        (rtc_numbers[0] > 23) ||
        (rtc_numbers[1] > 59) ||
        (rtc_numbers[2] > 59))
    {
        uart_write_string("Invalid time.\n");
        return;
    }

    rtc_time.hour = rtc_numbers[0];
    rtc_time.min = rtc_numbers[1];
    rtc_time.sec = rtc_numbers[2];

    rtc_set_time();
}

// If format is 1, prints the date time without formatting, if format is anything else, print date time with formatting.
void rtc_print_date_time(rtc_time_def *t, int format)
{
    rtc_get();

    if(format == 1){
        uart_write_hex_byte(uint8_to_bcd(t->month));
        uart_write_hex_byte(uint8_to_bcd(t->day));
        uart_write_hex_byte(uint8_to_bcd(t->year));
        uart_write_hex_byte(uint8_to_bcd(t->hour));
        uart_write_hex_byte(uint8_to_bcd(t->min));
        uart_write_hex_byte(uint8_to_bcd(t->sec));
    }
    else{
        uart_write_hex_byte(uint8_to_bcd(t->month));
        uart_write('/');
        uart_write_hex_byte(uint8_to_bcd(t->day));
        uart_write('/');
        uart_write_hex_byte(uint8_to_bcd(t->year));
        uart_write(' ');
        uart_write_hex_byte(uint8_to_bcd(t->hour));
        uart_write(':');
        uart_write_hex_byte(uint8_to_bcd(t->min));
        uart_write(':');
        uart_write_hex_byte(uint8_to_bcd(t->sec));
    }
}

// Converts RTC timestamp data to a more compact 4-byte format.
uint32_t rtc_date_to_timestamp(rtc_time_def *t)
{
    uint32_t ts;

    ts =
        (uint32_t)t->sec + 60L*((uint32_t)t->min + 60L*((uint32_t)t->hour + 24L*(((uint32_t)t->day-1) + 32L*(((uint32_t)t->month-1) + 12L*((uint32_t)t->year)))));

    return(ts);
}

void rtc_timestamp_to_date(uint32_t ts, rtc_time_def *t)
{
    uint32_t ts1;
    ldiv_t result;

    ts1 = ts;

    t->sec = ts1 % 60;
    ts1 = ts1 / 60;
        
    t->min = ts1 % 60;
    ts1 = ts1 / 60;
    
    t->hour = ts1 % 24;
    ts1 = ts1 / 24;

    t->day = ts1 % 32 + 1;
    ts1 = ts1 / 32;

    t->month = ts1 % 12 + 1;
    ts1 = ts1 / 12;

    t->year = ts1;
}

/*
void rtc_timestamp_to_date(uint32_t ts, rtc_time_def *t)
{
    uint32_t ts1;
    ldiv_t result;

    ts1 = ts;

    result = div_use(ts1, 60);
    t->sec = result.rem;
    ts1 = result.quot;
    
    result = div_use(ts1, 60);
    t->min = result.rem;
    ts1 = result.quot;

    result = div_use(ts1, 24);
    t->hour = result.rem;
    ts1 = result.quot;

    result = div_use(ts1,32);
    t->day = result.rem + 1;
    ts1 = result.quot;

    result = div_use(ts1, 12);
    t->month = result.rem + 1;
    ts1 = result.quot;

    t->year = ts1;
}
*/
