/*============================================================================
 * switch.c
 *      Modified to support two boards, where control pins are shared but
 *      the standby pin is unique for the two boards.
 * Author: Jon W. Wallace 
 *==========================================================================*/

#include <xc.h>
#include <inttypes.h>
#include "switch.h"
#include "global.h"

// Define antenna control ports
volatile uint8_t *ant_ports[MAX_PORTS] =
    {&PORTB, &PORTB, &PORTB, &PORTA, &PORTA, &PORTA, &PORTA, &PORTA,
     &PORTB, &PORTB, &PORTB, &PORTA};
volatile uint8_t *ant_lat[MAX_PORTS] =
    {&LATB, &LATB, &LATB, &LATA, &LATA, &LATA, &LATA, &LATA,
     &LATB, &LATB, &LATB, &LATA};
uint8_t ant_bits[MAX_PORTS] =
    {(1 << 3), (1 << 2), (1 << 1), (1 << 0), (1 << 1), (1 << 2), (1 << 3), (1 << 4),
     (1 << 3), (1 << 2), (1 << 1), (1 << 0)};
volatile uint8_t *ant_tris [MAX_PORTS] =
    {&TRISB, &TRISB, &TRISB, &TRISA, &TRISA, &TRISA, &TRISA, &TRISA,
     &TRISB, &TRISB, &TRISB, &TRISA};
volatile uint8_t *ant_ansel[MAX_PORTS] =
    {&ANSELB, &ANSELB, &ANSELB, &ANSELA, &ANSELA, &ANSELA, &ANSELA, &ANSELA,
     &ANSELB, &ANSELB, &ANSELB, &ANSELA};

// Keep track of previous antenna that was ON so we can easily turn it OFF.
static uint8_t last_idx;

void switch_init()
{
    uint8_t i;

    // Setup antenna ports as GPIO outputs and turn off.
    for (i=0; i<MAX_PORTS; i++)
    {
        *(ant_lat[i]) &= ~(ant_bits[i]);
        *(ant_ansel[i]) &= ~(ant_bits[i]);
        *(ant_tris[i]) &= ~(ant_bits[i]);
    }

    last_idx = 0xff;
}

// switch_set()
//      Sets specified switch for antenna idx to be ON.  If code 0xff is passed in, all antennas
//      are turned OFF.
void switch_set(uint8_t idx)
{
    // Turn last antenna OFF
    if (last_idx != 0xff)
    {
        *(ant_lat[last_idx]) &= ~(ant_bits[last_idx]);
        last_idx = 0xff;
    }

    if ( /*(idx >= 0) &&*/ (idx < MAX_ANTENNAS))
    {
        *(ant_lat[idx]) |= ant_bits[idx];
        last_idx = idx;
        
    }
}

void switch_sleep()
{
    uint8_t i;
    
    // Make sure antenna outputs are all off
    for (i=0; i<MAX_PORTS; i++)
    {
        *(ant_lat[i]) &= ~(ant_bits[i]);
    }

    last_idx = 0xff;
}

    
      
