/*============================================================================
 * tag_log.c
 *      Provides code to store tags in EEPROM.  A caching mechanism is used
 *      to efficiently store tags and not use much EEPROM space for
 *      repeated tags.
 * Author: Jon W. Wallace 
 *==========================================================================*/

#include <string.h>

#include "tag_log.h"

#include "rtc_timer.h"
#include "timer_ms.h"
#include "rfid.h"
#include "uart.h"
#include "eeprom.h"
#include "config.h"
#include "switch.h"
#include "battery.h"
#include "main.h"
#include "command.h"

#define LOG_POINTER_UNUSED      0xffff

//#define DEBUG_CACHE

#define TAG_MATCH(tag1, tag2)   ((tag1[0] == tag2[0]) && (tag1[1] == tag2[1]) && (tag1[2] == tag2[2]) && (tag1[3] == tag2[3]) && (tag1[4] == tag2[4]))

uint8_t tag_code[5] = {0xaa, 0xaa, 0xaa, 0xaa, 0xaa};

log_index_def log_index;                // Current position in the EEPROM where the log pointer is stored
log_pointer_def log_pointer;            // Log pointer value

tag_cache_entry_def tag_cache[MAX_CACHE_ENTRIES];
log_entry_def log_entry;

void tag_log_cache_clear()
{
    // Initialize cache
    memset(tag_cache, 0, sizeof(tag_cache));
}

void tag_log_init()
{
    // Read memory to find last log pointer
    if (tag_log_get_pointer(&log_index, &log_pointer) != 0)
    {
        uart_write_string("Error searching for log pointer.  Setting to 0.\r\n");
        log_index = 0;
        log_pointer = 0;
    }
    tag_log_cache_clear();

}

int8_t tag_log_erase_log_pointers()
{
    uint32_t i;
    uint8_t byte;
    uint8_t write_byte;
    
    // Set tag pointer memory in eeprom to all ffs.
    write_byte = 0xff;
    for (i=0; i<EEPROM_LOG_POINTER_SIZE; i++)
    {
        if (eeprom_read_bytes(EEPROM_LOG_POINTER_BASE+i, &byte, 1) != 0)
        {
            uart_write_string("Log ptr erase: EEPROM read error.\r\n");
            return(-1);
        }
        if ((byte != 0xff) && (eeprom_write_bytes(EEPROM_LOG_POINTER_BASE+i, &write_byte, 1) != 0))
        {
            uart_write_string("Log ptr erase: EEPROM write error.\r\n");
            return(-1);
        }
    }

    log_index = 0;
    log_pointer = 0;
    return(0);
}

// tag_log_get_pointer
//      Reads EEPROM to find a pointer to the end of the tag log.   We
//      use a "scratch pad" method where the pointer is stored in
//      successive memory locations rather than just at one address.
//      This way we avoid writing to the same location over and over
//      which wears out the EEPROM.
int8_t tag_log_get_pointer(log_index_def *found_idx, log_pointer_def *found_val)
{
    log_pointer_def last_val, curr_val;
    log_index_def   last_idx, curr_idx;

    uart_write_string("Scanning tag log...\r\n");
    
    // Search either for a place where the pointer stops ascending
    // or FFFF is found.  Either case indicates the end of the
    // pointer list.

    // Read first entry
    last_val = 0;          // Lowest possible index
    last_idx = 0;
    for (curr_idx=0; curr_idx<EEPROM_LOG_POINTER_SIZE/sizeof(log_pointer_def); curr_idx++)
    {
        if(eeprom_read_bytes(
                             LOG_POINTER_ADDRESS(curr_idx),
                             (uint8_t *) &curr_val,
                             sizeof(log_pointer_def)) != 0)
        {
            return(-1);
        }

        /* DEBUG
        uart_write_dec_int(last_idx);
        uart_write(' ');
        uart_write_dec_int(last_val);
        uart_write(' ');
        uart_write_dec_int(curr_idx);
        uart_write(' ');
        uart_write_dec_int(curr_val);
        uart_write_nl();
        */
        
        if ((curr_val < last_val) || (curr_val == LOG_POINTER_UNUSED))
        {
            *found_idx = last_idx;
            *found_val = last_val;
            return(0);
        }
        
        last_idx = curr_idx;
        last_val = curr_val;
    }

    // Got to the end without a match.  Last entry of the table must be current one.
    *found_idx = last_idx;
    *found_val = last_val;
    
    return(0);
}

// tag_log_store_entry_eeprom()
//     Stores entry in log_entry structure to
//     the EEPROM and updates pointers as needed.
int8_t tag_log_store_entry_eeprom()
{
    int8_t ccode;
    
    if (log_pointer >= LOG_ENTRY_MAX)
    {
        uart_write_string("Log full!\r\n");
        return(-1);
    }
    
    if ((ccode = eeprom_write_bytes(LOG_ENTRY_ADDRESS(log_pointer), (uint8_t *) &log_entry, sizeof(log_entry))) != 0)
    {
        uart_write_string("EEPROM log write error: ");
        uart_write_hex_byte((uint8_t) ccode);
        uart_write_nl();
        return(-2);
    }
    
    log_pointer++;
    
    // Store updated log pointer in EERPOM.  Store in a circular buffer in EEPROM to
    // extend life of EEPROM.
    if ((ccode = eeprom_write_bytes(LOG_POINTER_ADDRESS(log_index), (uint8_t *) &log_pointer, sizeof(log_pointer))) != 0)
    {
        uart_write_string("EEPROM log pointer write error: ");
        uart_write_hex_byte((uint8_t) ccode);
        uart_write_nl();
        return(-3);
    }
    
    log_index++;
    if (log_index >= LOG_POINTER_MAX) log_index = 0;

    return(0);
}

void tag_log_cache_entry_to_log_entry(tag_cache_entry_def *t)
{
    uint8_t i;
    uint16_t ant_mask;
    uint16_t ant_bit;
    
    log_entry.time_stamp = t->start_time_stamp;
    memcpy(log_entry.rfid, t->rfid, sizeof(log_entry.rfid));
    ant_mask = 0;
    ant_bit = 1;
    for (i=0; i<MAX_ANTENNAS; i++)
    {
        if (t->antenna_hits[i]) ant_mask |= ant_bit;
        ant_bit <<= 1;
    }
    log_entry.antenna_mask = ant_mask;
}

// tag_log_store()
//      This function stores a log entry to the cache / log.  Pass in 0xff for the antenna
//      index to just check for entries that have timed out and need to go to EEPROM.
int8_t tag_log_store(uint8_t ant_idx, uint8_t *rfid, uint32_t time_stamp)
{
    uint8_t empty_found;
    uint8_t do_store_cache;
    uint32_t dt_oldest;
    uint32_t dt;
    uint32_t tnow;
    uint8_t i;
    tag_cache_entry_def *t, *t_oldest;
#ifdef DEBUG_CACHE
    uart_write_string("Cache: tag_log_store()\r\n");
#endif

    tnow = timer_ms_read();
    
    if (system_config.cache_size == 0)
    {
#ifdef DEBUG_CACHE
        uart_write_string("Cache: Direct write.\r\n");
#endif
        // No cache.  Write directly to log (if not just checking)
        if (ant_idx == 0xff) return(0);
        memcpy(log_entry.rfid, rfid, sizeof(log_entry.rfid));
        log_entry.time_stamp = time_stamp;
        log_entry.antenna_mask = 1 << ant_idx;

        tag_log_store_entry_eeprom();
    }
    else
    {
        // Check for expired cache entries, writing to EEPROM as needed.
        t = tag_cache;
        for (i=0; (i<system_config.cache_size); i++, t++)
        {
            if (t->valid)
            {
                dt = tnow - t->start_time_ms;
                if (dt >= (uint32_t)system_config.cache_timeout*1000L)
                {
#ifdef DEBUG_CACHE
                    uart_write_string("Cache: (expired)\r\n");
#endif
                    // Expired.  Write the entry to the log.
                    tag_log_cache_entry_to_log_entry(t);
#ifdef DEBUG_CACHE
                    uart_write_string("Cache: Writing EEPROM\r\n");
#endif        
                    tag_log_store_entry_eeprom();

                    // Clear out entry in cache.
                    memset(t, 0, sizeof(tag_cache_entry_def));
                }
            }
        }
        // If we don't have a real tag entry to store, we are done.
        if (ant_idx == 0xff) return(0); 
        
        // Try to find entry in cache.  Also, try to find an unused entry or otherwise
        // the oldest entry in case it is not there yet.
        empty_found = 0;
        dt_oldest = 0;
        t_oldest = 0;
        t = tag_cache;
        do_store_cache = 1;
        for (i=0; (do_store_cache) && (i<system_config.cache_size); i++, t++)
        {
            if (!empty_found && !(t->valid))
            {
                // We don't already have an unused entry yet.
#ifdef DEBUG_CACHE
                uart_write_string("Cache: Unused entry ");
                uart_write_dec_int((uint16_t) (t-tag_cache));
                uart_write_nl();
#endif
                // Unused entry. Make this look as old as possible
                empty_found = 1;
                dt_oldest = 0xffffffff;
                t_oldest = t;
            }
            else if (TAG_MATCH(t->rfid, rfid))
            {
#ifdef DEBUG_CACHE
                uart_write_string("Cache: ID matches entry ");
                uart_write_dec_int((uint16_t) (t-tag_cache));
                uart_write_nl();
#endif
                // Entry is updated in place, so do not store again.
                do_store_cache = 0;

                // Assume tag has not expired.  Update number of hits.
                if (t->antenna_hits[ant_idx] != 0xffff) t->antenna_hits[ant_idx]++;
            }
            else 
            {
                // Have a valid entry that is not a match.  Keep track of it if it
                // is older than the one we currently are keeping track of.
                dt = (tnow - t->start_time_stamp);  // stores age in ms.
                if (dt > dt_oldest)
                {
                    // Entry is valid and older than previous one found
                    dt_oldest = dt;
                    t_oldest = t;
                }
            }
        }

        if (do_store_cache)
        {
#ifdef DEBUG_CACHE
            uart_write_string("Cache: Overwriting unused/oldest entry ");
            uart_write_dec_int(t_oldest - tag_cache);
            uart_write_nl();
#endif
            // Store the one being overwritten (if valid) to EEPROM
            if (t_oldest->valid)
            {
                tag_log_cache_entry_to_log_entry(t_oldest);
#ifdef DEBUG_CACHE
                uart_write_string("Cache: Writing EEPROM\r\n");
#endif        
                tag_log_store_entry_eeprom();
            }
            
            // Store the RFID in unused slot or oldest one found
            t_oldest->valid = 1;
            memcpy(t_oldest->rfid, rfid, sizeof(t_oldest->rfid));
            t_oldest->start_time_stamp = time_stamp;
            t_oldest->start_time_ms = timer_ms_read();
            memset(t_oldest->antenna_hits, 0, sizeof(t_oldest->antenna_hits));
            t_oldest->antenna_hits[ant_idx] = 1;
        }
    }

    return(0);
}

// Writes tags that are in the tag cache to EEPROM and clears cache.
void tag_log_flush()
{
    uint8_t i;
    tag_cache_entry_def *t;

    t = tag_cache;
    for (i=0; i<system_config.cache_size; i++, t++)
    {
        if (t->valid)
        {
            tag_log_cache_entry_to_log_entry(t);
            tag_log_store_entry_eeprom();
        }
    }

    tag_log_cache_clear();
}

// tag_log_erase()
//      This function erases the contents of the tag log and resets to empty state.
//
int8_t tag_log_erase()
{
    // Zero out the cache
    memset(tag_cache, 0, sizeof(tag_cache));

    // Erasing the pointers in EEPROM and setting current pointer to 0 is
    // all we need to do.
    if (tag_log_erase_log_pointers() != 0)
    {
        return(-1);
    }

    return(0);
}

rtc_time_def log_time;
int8_t tag_log_dump()
{
    log_pointer_def log_pointer1;
    uint32_t usage;
    uint8_t i;

    // Print some stats of the log
    usage = (uint32_t)log_pointer*1000L/LOG_ENTRY_MAX;
    uart_write_string("Log usage: ");
    uart_write_dec_int(log_pointer);
    uart_write('/');
    uart_write_dec_int(LOG_ENTRY_MAX);
    uart_write_string(" entries (");
    uart_write_dec_int(usage/10);
    uart_write('.');
    uart_write_dec_int(usage%10);
    uart_write_string("%)\r\n");

    // Print the contents of the tag log
    uart_write_string("Log START\r\n");

    // Reset CRC
    uart_crc_reset();
    uart_crc_on();
    
    for (log_pointer1=0; log_pointer1 < log_pointer; log_pointer1++)
    {
        // Read the entry
        if (eeprom_read_bytes(LOG_ENTRY_ADDRESS(log_pointer1), (uint8_t *) &log_entry, sizeof(log_entry)) != 0)
        {
            uart_write_string("EEPROM read error.\r\n");
            uart_crc_off();
            return(-1);
        }

        rtc_timestamp_to_date(log_entry.time_stamp, &log_time);
        rtc_print_date_time(&log_time, 1);

        for (i=0; i<5; i++)
        {
            uart_write_hex_byte(log_entry.rfid[i]);
        }

        uart_write_hex_byte(log_entry.antenna_mask >> 8);
        uart_write_hex_byte(log_entry.antenna_mask & 0xff);
        
        uart_crc_off();
        uart_write_nl();
        uart_crc_on();
        
        if (uart_check_break())
        {
            uart_crc_off();
            uart_write_string("Dump terminated.\r\n");
            return(0);
        }
        
    }

    uart_crc_off();

    uart_write_string("CRC: ");
    uart_write_crc();
    uart_write_string("----------------------\r\nLog END--------------------\r\n");
    
    return(0);
}

int8_t tag_log_dump_formatted()
{
    log_pointer_def log_pointer1;
    uint32_t usage;
    uint8_t i;

    // Print some stats of the log
    usage = (uint32_t)log_pointer*1000L/LOG_ENTRY_MAX;
    uart_write_string("Log usage: ");
    uart_write_dec_int(log_pointer);
    uart_write('/');
    uart_write_dec_int(LOG_ENTRY_MAX);
    uart_write_string(" entries (");
    uart_write_dec_int(usage/10);
    uart_write('.');
    uart_write_dec_int(usage%10);
    uart_write_string("%)\r\n");

    // Print the contents of the tag log
    uart_write_string("Log START\r\n");

    // Reset CRC
    uart_crc_reset();
    uart_crc_on();
    
    for (log_pointer1=0; log_pointer1 < log_pointer; log_pointer1++)
    {
        // Read the entry
        if (eeprom_read_bytes(LOG_ENTRY_ADDRESS(log_pointer1), (uint8_t *) &log_entry, sizeof(log_entry)) != 0)
        {
            uart_write_string("EEPROM read error.\r\n");
            uart_crc_off();
            return(-1);
        }

        rtc_timestamp_to_date(log_entry.time_stamp, &log_time);
        rtc_print_date_time(&log_time, 0);
        uart_write(' ');

        for (i=0; i<5; i++)
        {
            uart_write_hex_byte(log_entry.rfid[i]);
        }

        uart_write(' ');
        uart_write_hex_byte(log_entry.antenna_mask >> 8);
        uart_write_hex_byte(log_entry.antenna_mask & 0xff);
        
        uart_crc_off();
        uart_write_nl();
        uart_crc_on();
        
        if (uart_check_break())
        {
            uart_crc_off();
            uart_write_string("Dump terminated.\r\n");
            return(0);
        }
        
    }

    uart_crc_off();

    uart_write_string("CRC: ");
    uart_write_crc();
    uart_write_string("----------------------\r\nLog END--------------------\r\n");
    
    return(0);
}


// tag_log_run()
//      Polls for RFID tags and stores in the log.  Can stop with [ESC].
void tag_log_run()
{
    int8_t ccode, quit;
    uint32_t t1, t2, dt;
    uint32_t ts;
    uint8_t ant_idx;

    tag_log_cache_clear();
    
    uart_write_string("Press "COMMAND_BREAK_STR" to stop:\r\n");

    ant_idx = system_config.ant_start;
    switch_set(ant_idx);

    quit = 0;
    for ( ; ; )
    {
        uart_write_hex_byte(ant_idx);
        uart_write(':');
        uart_write(' ');
        
        ccode = rfid_read(system_config.time_poll, ant_idx);
        t1 = timer_ms_read();
        switch (ccode)
        {
            case RFID_READ_TIMEOUT:
                uart_write_string("FFFFFFFFFF");

                // Just check tag cache for expired entries.
                tag_log_store(0xff, 0, 0);
                break;
                
            case RFID_READ_OK:
                // Print tag info
                rfid_print_data();

                // Get current time and compute timestamp
                rtc_get();
                ts = rtc_date_to_timestamp(&rtc_time);

                // Store
                tag_log_store(ant_idx, rfid_data_out, ts);
                break;
        }

        uart_write(' ');
        battery_main_print();
        uart_write_nl();

        if (uart_check_break()) break;

        // Wait half the delay time, switch, and then again half the delay time

        //t1 = timer_ms_read();  // Read moved above so that this delay time can be overlaid with checking cache
        for ( ; ; )
        {
            t2 = timer_ms_read();
            dt = t2-t1;
            if (dt >= (system_config.time_delay >> 1))
            {
                break;
            }

            if (uart_check_break())
            {
                quit = 1;
                break;
            }
        }
        if (quit) break;
        
        if (system_config.ant_stop > system_config.ant_start)
        {
            // Ascending range
            ant_idx++;
            if (ant_idx > system_config.ant_stop) ant_idx = system_config.ant_start;
            switch_set(ant_idx);
        }
        else if (system_config.ant_stop < system_config.ant_start)
        {
            // Descending range
            ant_idx--;
            if ((ant_idx == 0xff) || (ant_idx < system_config.ant_start)) ant_idx = system_config.ant_stop;
            switch_set(ant_idx);
        }

        t1 = timer_ms_read();
        for ( ; ; )
        {
            t2 = timer_ms_read();
            dt = t2-t1;
            if (dt >= (system_config.time_delay >> 1))
            {
                break;
            }

            if ((uart_read_ready()) && (uart_read() == COMMAND_BREAK_CHAR))
            {
                quit = 1;
                break;
            }

        }
        if (quit) break;

        // Check battery
        if (battery_main_mv() < system_config.batt_off_threshold)
        {
            uart_write_string("Battery below threshold.  Entering sleep.\r\n");
            if (system_config.cache_size) tag_log_flush();

            // Write a code to indicate sleep
            tag_log_write_code(TAG_LOG_CODE_SLEEP);
            
            timer_ms_delay(100);

            // Sleep and never return.
            do_sleep();
        }
        
    }

    switch_set(SWITCH_ALL_OFF);
    
    if (system_config.cache_size) tag_log_flush();
}

// tag_log_write_code
//   Writes a specific code to the eeprom.  Used to indicate events
//   like power-up, sleep, wake-up.
void tag_log_write_code(uint8_t sub_code)
{
    uint32_t ts;
    
    log_entry.time_stamp = rtc_date_to_timestamp(&rtc_time);
    memcpy(log_entry.rfid, tag_code, sizeof(log_entry.rfid));
    log_entry.antenna_mask = sub_code;
    tag_log_store_entry_eeprom();
}





