/*============================================================================
 * timer.c
 *      Code to use timer0 for higher precision timing than is possible with
 *      the real-time clock timer.  The prescaler needs to be >= 4
 *      for it to work.
 *
 * Important note:
 *      Usually we use this as an unsigned 8-bit timer that is constantly
 *      running.  When checking for elapsed time, we usually get the time in
 *      t1, later read the time again as t2, and check t2-t1 > interval.  For
 *      some reason, this comparison does not handle the rollover directly,
 *      but it does work if you place t2-t1 first in a uint8 and then
 *      do the check.  It seems like this should be equivalent, but the
 *      XC8 compiler apparently being too smart to allow it to work.
 *
 * Author: Jon W. Wallace 
 *==========================================================================*/

#include <xc.h>
#include <inttypes.h>

#include "timer.h"

#define T0CON0_T016BIT  (1 << 4)
#define T0CON0_T0EN     (1 << 7)

#define T0CON1_T0CS       (0x3 << 5)      // HFINTOSC
#define T0CON1_T0CKPS_1   TIMER_PRESCALER_POW

void timer_init(void)
{
    // Make sure timer is off and set to be a 16-bit timer.
#ifdef TIMER_16BIT
    T0CON0 = T0CON0_T016BIT;
#else
    T0CON0 = 0;
#endif
    
    // Initialize timer to 0.  High then low byte.
#ifdef TIMER_16BIT    
    TMR0H = 0;
#else
    TMR0H = 0xff;
#endif    
    TMR0L = 0;

    // Set up prescaler and clock source
    T0CON1 = T0CON1_T0CS | T0CON1_T0CKPS_1;
}

void timer_reset(void)
{
#ifdef TIMER_16BIT    
    TMR0H = 0;
#else
    TMR0H = 0xff;
#endif    
    TMR0L = 0;
}

void timer_on(void)
{
    T0CON0 |= T0CON0_T0EN;
    timer_reset();
}

void timer_off(void)
{
    T0CON0 &= ~T0CON0_T0EN;
}

timer_t timer_read(void)
{
    uint8_t tmrh, tmrl;
    
    // Read low first, which latches high part simultaneously.
#ifdef TIMER_16BIT
    tmrl = TMR0L;
    tmrh = TMR0H;
    return((uint16_t)tmrl | ((uint16_t)tmrh << 8));
#else
    return(TMR0L);
#endif
}

void timer_delay_ticks(timer_t tics)
{
    timer_t t1, t2, dt;

    t1 = TIMER_READ;
    do
    {
        t2 = TIMER_READ;
        dt = t2-t1;
    } while (dt < tics);
}

void timer_sleep()
{
    // Turn the timer OFF
    T0CON0 = 0;
    T0CON1 = 0;
}




