/*============================================================================
 * timer_ms.c
 *	Module for creating a 32-bit reusable millisecond timer.
 *
 * Note: Timer will roll over once every 50 days.
 *
 * Author: Jon Wallace
 *==========================================================================*/

#include <xc.h>
#include <inttypes.h>
#include "mcc_generated_files/mcc.h"
#include "global.h"

#define CLOCK_FREQ (_XTAL_FREQ)

#define TxCLKCON_HFINTOSC       0x03

#define TxCON_ON                (1 << 7)
#define TxCON_CKPS_8            (0x3 << 4)
#define TxCON_OUTPS_10          (0x9)

#define PIE4_TMR2IE             (1 << 1)
#define PIR4_TMR2IF             (1 << 1)

volatile uint32_t timer_ms_count;

void timer_ms_init()
{
    // Reset our global counter variable
    timer_ms_count = 0;

    // Use high frequency input clock
    T2CLKCON = TxCLKCON_HFINTOSC;

    // Set up pre-post scalers for /80
    T2CON = TxCON_OUTPS_10 | TxCON_CKPS_8;

    // Reset the timer value
    TMR0 = 0;
    
    // Periodic reset value
    T2PR = CLOCK_FREQ/80/1000;
    
    // Enable the interrupt
    PIE4 |= PIE4_TMR2IE;

    // Turn on the timer
    T2CON |= TxCON_ON;
}

inline void timer_ms_isr()
{
    //RFID_DEBUG_LAT ^= RFID_DEBUG_BIT;
    
    // Check if we caused the interrupt
    // if (PIR4 & PIR4_TMR2IF) // It is the only interrupt source 
    {
        // Bit is set.  Increment our counter and clear the flag.
        timer_ms_count++;
        PIR4 &= ~PIR4_TMR2IF;
    }
}

uint32_t timer_ms_read()
{
    uint32_t t1;

    // Turn off interrupts to make read of 32-bit counter atomic
    GINT_OFF;
    t1 = timer_ms_count;
    GINT_ON;
    return(t1);
}

uint16_t timer_ms_read_uint16()
{
    uint16_t t1;

    // Turn off interrupts to make read of 32-bit counter atomic
    GINT_OFF;
    t1 = (uint16_t)timer_ms_count;
    GINT_ON;
    return(t1);
}

void timer_ms_delay(uint32_t ms)
{
    uint32_t t1, dt;

    t1 = timer_ms_read();
    for ( ; ; )
    {
        dt = timer_ms_read() - t1;
        if (dt >= ms) break;
    }
}

void timer_ms_sleep()
{
    // Disnable the timer interrupt
    PIE4 &= ~PIE4_TMR2IE;

    // Turn off the timer
    T2CON = 0;
}

