/*============================================================================
 * uart.c
 *	UART library.
 * Author: Jon W. Wallace 
 * Notes:
 *      Uses EUSART1 on pins RC6 (TX pin 17) and RC7 (RX pin 18)
 *==========================================================================*/

#include <xc.h>
#include <inttypes.h>

#include "mcc_generated_files/mcc.h"
#include "uart.h"
#include "command.h"
#include "timer_ms.h"

// Redfine registers to allow easy switching between two serial ports
// UART1 (RC6=TX 16 RC7=RX)
#define RXxPPS          RX1PPS
#define RXxPPS_VAL      0x17    // Use RC7 for input for Rx1
#define RxxPPS          RC6PPS
#define RxxPPS_VAL      0x0d    // Use RC6 as output for Tx1
#define TXxSTA          TX1STA
#define RCxSTA          RC1STA
#define BAUDxCON        BAUD1CON
#define SPxBRGL         SP1BRGL
#define SPxBRGH         SP1BRGH
#define TXxREG          TX1REG
#define RCxREG          RC1REG
#define UART_PORT       (PORTC)
#define UART_LAT        (LATC)
#define UART_PORT_ANSEL (ANSELC)
#define UART_PORT_TRIS  (TRISC)
#define UART_PORT_WPU   (WPUC)
#define PMD4_UARTxMD    (1 << 6)                // Enable register for hardware module
#define PIR3_RCxIF      (1 << 5)                // Interrupt flag in PIR3 for receive

/*
// UART2 (RB6=TX RB7=RX)
#define TXxSTA          TX2STA
#define RCxSTA          RC2STA
#define BAUDxCON        BAUD2CON
#define SPxBRGL         SP2BRGL
#define SPxBRGH         SP2BRGH
#define TXxREG          TX2REG
#define RCxREG          RC2REG
#define UART_PORT_ANSEL         (ANSELB)
#define UART_PORT_TRIS          (TRISCB)
#define PMD4_UARTxMD    (1 << 7)
#define PIR3_RCxIF      (7 << 5)
*/

#define UART_PORT_TX            (1 << 6)
#define UART_PORT_RX            (1 << 7)

#define TXxSTA_CSRC             (1 << 7)
#define TXxSTA_TXEN             (1 << 5)
#define TXxSTA_SYNC             (1 << 4)
#define TXxSTA_BRGH             (1 << 2)
#define TXxSTA_TRMT             (1 << 1)

#define RCxSTA_SPEN             (1 << 7)
#define RCxSTA_CREN             (1 << 4)

// Power input pin to BT module
#define BT_VDD_PORT         (PORTA)
#define BT_VDD_LAT          (LATA)
#define BT_VDD_TRIS         (TRISA)
#define BT_VDD_ANSEL        (ANSELA)
#define BT_VDD_BIT          (1 << 6) // RA6
    
// Baud rate control
#define BAUDxCON_RCIDL          (1 << 6)
#define BAUDxCON_SCKP_LOW       (1 << 4)
#define BAUDxCON_BRG16          (1 << 3)
#define BAUDxCON_WUE            (1 << 1)

#define PCLK                    (_XTAL_FREQ)
#define BAUD_DESIRED            (19200)
//#define BAUD_DIV                64      // 8-bit, BRGH=0
//#define BAUD_DIV                16      // 16-bit, BRGH=0
//#define BAUD_DIV                16      // 8-bit, BRGH=1
#define BAUD_DIV                4       // 16-bit, BRGH=1
#define BAUD_REG                (PCLK/BAUD_DESIRED/(BAUD_DIV)-1)  // 8-bit mode, BRGH=1
#define BAUD_REG_L              (BAUD_REG & 0xff)
#define BAUD_REG_H              ((BAUD_REG >> 8) & 0xff)

// Key assignments

#define UART_AUTO_CR_LF   1
#define KEY_BS            8

uint8_t uart_mode = UART_MODE_UNKNOWN;

uint32_t uart_timeout_t1;
uint32_t uart_timeout_ms;

static uint16_t uart_crc;
static uint8_t uart_crc_flag;

// uart_set_timeout
//      Sets a timeout in ms that is honored by UART input routines.
void uart_set_timeout(uint32_t timeout_val)
{
    uart_timeout_t1 = timer_ms_read();
    uart_timeout_ms = timeout_val;
}

// uart_sense_jumper()
//      Checks to see if the mode jumper is set for RS232 (RA7 connected to RX port RC7).  In this
//      case, invert the serial port signals.  Otherwise, set port for bluetooth (uninverted signals).
//      This function just senses the jumper position.  The init function sets things up.
void uart_sense_jumper()
{
    uint16_t i;

    // Default mode is bluetooth (not inverted)
    uart_mode = UART_MODE_BT;
    
    // Set Rx pin to be a normal digital input
    UART_PORT_ANSEL &= ~UART_PORT_RX;
    UART_PORT_TRIS |= UART_PORT_RX;

    // Let a little bit of time pass
    for (i=0; i<16384; i++) NOP();
    
    // RS232 inverter has a pull-up and BT module is OFF, so for
    // RS232 we will see HIGH and for BT we will see LOW.
    if ((UART_PORT & UART_PORT_RX) == 0)
    {
        uart_mode = UART_MODE_BT;
    }
    else
    {
        uart_mode = UART_MODE_RS232;
    }

    // Turn off any values that might adversely affect use of pins later
    // UART_PORT_WPU  &= ~UART_PORT_RX;
}

// UART initialization
void uart_init()
{
    // Do not compute CRC of uart output by default.
    uart_crc_flag = 0;
    
    // Reset timeout values
    uart_timeout_t1 = 0;
    uart_timeout_ms = 0;

    // Turn off power to BT initially.  Needed to sense jumper.
    BT_VDD_LAT &= ~BT_VDD_BIT;         
    BT_VDD_ANSEL &= ~BT_VDD_BIT;        // Turn off analog
    BT_VDD_TRIS &= ~BT_VDD_BIT;         // Set to be output
    
    // Sense the UART mode position
    uart_sense_jumper();

    // For bluetooth mode, power up the bluetooth module
    if (uart_mode == UART_MODE_BT)
    {
        BT_VDD_LAT |= BT_VDD_BIT;
    }
    
    // Map peripheral pins to digital pins
    RXxPPS = RXxPPS_VAL;
    RxxPPS = RxxPPS_VAL;
    
    // Make UART peripheral active (clear the bit).  Already done by default.
    //PMD4 &= ~PMD4_UARTxMD;

    // Make sure analog functionality off on these pins
    UART_PORT_ANSEL &= ~(UART_PORT_TX | UART_PORT_RX);
    // Make input just in case
    UART_PORT_TRIS |= (UART_PORT_TX | UART_PORT_RX);

    // Set the baud rate
    SPxBRGH = BAUD_REG_H;
    SPxBRGL = BAUD_REG_L;
    TXxSTA = TXxSTA_BRGH;       // Set baud rate divider = 16, TX off
    RCxSTA = 0;
    BAUDxCON = BAUDxCON_BRG16;  // Use 16-bit timer and defaults
    if (uart_mode == UART_MODE_RS232)
    {
        BAUDxCON |= BAUDxCON_SCKP_LOW;  // Use 16-bit timer and defaults    
    }
    
    // Turn ON Rx
    RCxSTA |= (RCxSTA_SPEN | RCxSTA_CREN);
    
    // Turn ON Tx
    TXxSTA |= TXxSTA_TXEN;

    //uart_write_string("UART_MODE=");
    //uart_write_hex_byte(uart_mode);
    //uart_write_nl();
}

#define CRC16_POLY  0x1021
void do_crc(uart_char c)
{
    uint8_t j;

    for (j = 0; j < 8; j++)
    {
        if ((c >> 7) ^ (uart_crc >> 15))
        {
            uart_crc <<= 1;
            uart_crc ^= CRC16_POLY;
        }
        else
        {
            uart_crc <<= 1;
        }
        c <<= 1;
    }
}

void uart_write_crc()
{
    uart_write_hex_byte((uart_crc >> 8) & 0xff);
    uart_write_hex_byte(uart_crc & 0xff);
}

void uart_crc_on()
{
    uart_crc_flag = 1;
}

void uart_crc_off()
{
    uart_crc_flag = 0;
}

void uart_crc_reset()
{
    uart_crc = 0;
}

// Blocking write
void uart_write(uart_char ch)
{
    // Wait for Tx buffer to not be full
    while (!(TXxSTA & TXxSTA_TRMT));

    // Put in FIFO
    TXxREG = ch;

    // Compute CRC if desired
    if (uart_crc_flag) do_crc(ch);
}

void uart_write_nl(void)
{
    uart_write('\r');
    uart_write('\n');
}

// Block write string function
void uart_write_string(uart_char *s)
{
    while (*s)
    {
#ifdef UART_AUTO_CR_LF
        if (*s == '\n') uart_write('\r');
#endif
        uart_write(*s);
        s++;
    }
}

// Uart_write_prompt()
//      Outputs a prompt.  When we are in bluetooth mode,
//      puts an extra newline after the prompt so the prompt will
//      be displayed.
void uart_write_prompt(uart_char *s)
{
    uart_write_string(s);
    if (uart_mode == UART_MODE_BT)
    {
        uart_write_nl();
    }
}

uart_char hex_to_char(uint8_t n)
{
    if (n <= 0x9)
    {
        return(n + '0');
    }
    else if (n <= 0xf)
    {
        return(n-0xa+'A');
    }
    else
    {
        return('?');
    }
}

void uart_write_hex_byte(uint8_t v)
{
    uart_write(hex_to_char((v >> 4) & 0xf));
    uart_write(hex_to_char(v & 0xf));
}

void uart_write_hex_long(uint32_t v)
{
    uint8_t i;

    for (i=0; i<8; i++)
    {
        uart_write(hex_to_char(v >> 28));
        v <<= 4;
    }
}

void uart_write_dec_int(uint32_t n)
{
    uint32_t mult1;
    uint8_t dig;
    uint8_t i;
    
    if (n == 0)
    {
        uart_write('0');
        return;
    }

    mult1 = 1;
    for (i=0; i<=9; i++)
    {
        if ((n/mult1) == 0)
        {
            mult1 /= 10;
            break;
        }
        else if (i != 9)
        {
            mult1 *= 10;
        }
    }

    //uart_write_hex_long(mult1);
    //uart_write_nl();
    
    while(mult1)
    {
        dig = n/mult1;
        uart_write(dig + '0');
        n -= dig*mult1;
        mult1 /= 10;
    }
}

// Block read string function.  Reads until a cr '\r' is seen.
// eats any newlines '\n' that are seen.
void uart_read_string(uart_char *s, uint32_t buffer_size, uart_char echo)
{
    uart_char c;
    uint32_t idx;
    uart_char done;
    
    idx = 0;
    for ( ; ; )
    {
        c = uart_read();
        switch(c)
        {
            case UART_TIMEOUT_CHAR: 
                s[0] = UART_TIMEOUT_CHAR;
                s[1] = 0;
                return;
                
            case KEY_BS: // backspace
                if (idx > 0)
                {
                    idx--;
                    if (echo)
                    {
                        uart_write(KEY_BS);
                        uart_write(' ');
                        uart_write(KEY_BS);
                    }
                }
                break;
            case '\n':
                continue;
                break;
            case '\r':
                // Terminate string and end
                s[idx] = 0;
                if (echo)
                {
#ifdef UART_AUTO_CR_LF
                    uart_write('\r');
#endif            
                    uart_write('\n');
                }
                return;
            default:
                // Append character to string
                s[idx] = c;
                if (echo) uart_write(c);
                idx++;
                break;
        }

        // Check if we have reached buffer size (-1: need space for null also)
        if (idx >= buffer_size-1)
        {
            s[idx] = 0;
            return;
        }
    }
}

// Blocking read
uart_char uart_read()
{
    uint32_t dt;
    
    // Wait for character to be ready by polling interrupt flag.
    while (!(PIR3 & PIR3_RCxIF))
    {
        // Check and enforce timeout if enabled
        if (uart_timeout_ms != 0)
        {
            dt = timer_ms_read() - uart_timeout_t1;
            if (dt > uart_timeout_ms)
            {
                uart_timeout_t1 = 0;
                uart_timeout_ms = 0;
                return(UART_TIMEOUT_CHAR);
            }
        }
    }
           
    // Return the character
    return(RCxREG);
}

// Check write ready
uint8_t uart_write_ready()
{
    return (TXxSTA & TXxSTA_TRMT);
}

// Check read ready (character available in buffer)
uint8_t uart_read_ready()
{
    // Output TRUE if buffer is not full
    return ((PIR3 & PIR3_RCxIF) != 0);
}

void uart_sleep()
{
    // Turn OFF Rx
    RCxSTA &= ~(RCxSTA_SPEN | RCxSTA_CREN);
    
    // Turn OFF Tx
    TXxSTA &= ~TXxSTA_TXEN;

    // Turn off the bluetooth module
    BT_VDD_LAT &= ~BT_VDD_BIT;
}

// uart_check_break()
// Function to check for a break character.  We handle things slightly differently
// depending on whether it is an RS232 or Bluetooth connection.  For RS232, a single
// break character stops things.  For Bluetooth, we expect the break character followed
// by a carriage return.
#define UART_CHECK_BREAK_TIMEOUT_MS     500
uint8_t uart_check_break()
{
    uint32_t t1, dt;
    
    // Check for the break character
    if (uart_read_ready() && (uart_read() == COMMAND_BREAK_CHAR))
    {
        // If not in bluetooth mode, the break character alone is enough.
        if (uart_mode != UART_MODE_BT) return(1);
        
        // If in bluetooth mode, wait for a newline or timeout.
        t1 = timer_ms_read();
        for ( ; ; )
        {
            if (uart_read_ready() && (uart_read() == '\r'))
            {
                uart_flush(0, 0);  // Bluetooth gets messed up when extra junk in buffer.
                return(1);
            }
            dt = timer_ms_read() - t1;
            if (dt > UART_CHECK_BREAK_TIMEOUT_MS)
            {
                // Timeout waiting for CR.  Say that the break character was not received.
                return(0);
            }
        }
    }
    
    // Break not received.
    return(0);
}

// uart_flush()
//      Read any junk coming from the UART for a short time.  Optionally, store
//      data to a buffer (for debug).
#define UART_FLUSH_TIMEOUT     1000
uint16_t uart_flush(uart_char *s, uint32_t buffer_size)
{
    uint32_t t1, dt;
    uint16_t idx;
    uart_char c;
    
    t1 = timer_ms_read();
    idx = 0;
    for ( ; ; )
    {
        if (uart_read_ready())
        {
            c = uart_read();
            if (s)
            {
                if (idx < buffer_size)
                {
                    s[idx] = c;
                    idx++;
                }
            }
        }
        dt = timer_ms_read() - t1;
        if (dt > UART_FLUSH_TIMEOUT) break;
    }
    return(idx);
}

